#!/usr/bin/env python
import copy
import logging
from time import time
from typing import Callable

import numpy as np
from tqdm import tqdm

import mapel.core.persistence.experiment_exports as exports
from mapel.core.inner_distances import map_str_to_func
from mapel.core.objects.Experiment import Experiment
from mapel.elections.distances import main_approval_distances as mad
from mapel.elections.distances import main_ordinal_distances as mod
from mapel.elections.objects.ApprovalElection import ApprovalElection
from mapel.elections.objects.OrdinalElection import OrdinalElection

registered_approval_distances = {
    'approvalwise': mad.compute_approvalwise,

    'hamming': mad.compute_hamming,  # unsupported distance
}

registered_ordinal_distances = {
    'positionwise': mod.compute_positionwise_distance,
    'bordawise': mod.compute_bordawise_distance,
    'pairwise': mod.compute_pairwise_distance,
    'discrete': mod.compute_discrete_distance,

    'swap': mod.compute_swap_distance,
    'spearman': mod.compute_spearman_distance,

    'blank': mod.compute_blank_distance,

    'ilp_spearman': mod.compute_spearman_distance_ilp_py,  # unsupported distance
    'ilp_swap': mod.compute_swap_distance_ilp_py,  # unsupported distance
    'voterlikeness': mod.compute_voterlikeness_distance,  # unsupported distance
    'agg_voterlikeness': mod.compute_agg_voterlikeness_distance,  # unsupported distance
    'pos_swap': mod.compute_pos_swap_distance,  # unsupported distance
    'voter_subelection': mod.compute_voter_subelection,  # unsupported distance
    'candidate_subelection': mod.compute_candidate_subelection,  # unsupported distance
}


def add_approval_distance(name, function):
    """
    Adds a new approval distance to the list of approval distances.

    :param name: name of the distance.
    :param function: function that computes the distance.
    :return: None.
    """
    registered_approval_distances[name] = function


def add_ordinal_distance(name, function):
    """
    Adds a new ordinal distance to the list of ordinal distances.

    :param name: name of the distance.
    :param function: function that computes the distance.
    :return: None.
    """
    registered_ordinal_distances[name] = function


def get_distance(election_1,
                 election_2,
                 distance_id: str = None) -> float or (float, list):
    """
    Computes distance between elections, (if applicable) optimal matching.

    :param election_1: first election.
    :param election_2: second election.
    :param distance_id: name of the distance.
    :return: distances, matching (if applicable).
    """
    if type(election_1) is ApprovalElection and type(election_2) is ApprovalElection:
        return get_approval_distance(election_1, election_2, distance_id=distance_id)
    elif type(election_1) is OrdinalElection and type(election_2) is OrdinalElection:
        return get_ordinal_distance(election_1, election_2, distance_id=distance_id)
    else:
        logging.warning('No such instance!')


def get_approval_distance(election_1: ApprovalElection, election_2: ApprovalElection,
                          distance_id: str = None, **kwargs) -> (float, list):
    """
    Computes distance between approval elections, (if applicable) optimal matching.

    :param election_1: first election.
    :param election_2: second election.
    :param distance_id: name of the distance.
    :param kwargs: additional arguments.
    :return: distances, matching (if applicable).
    """

    inner_distance, main_distance = _extract_distance_id(distance_id)

    if main_distance in registered_approval_distances:

        if inner_distance is not None:
            return registered_approval_distances.get(main_distance)(election_1,
                                                                    election_2,
                                                                    inner_distance)
        else:
            return registered_approval_distances.get(main_distance)(election_1,
                                                                    election_2,
                                                                    **kwargs)

    else:
        logging.warning(f'No such distance as: {main_distance}!')


def get_ordinal_distance(election_1: OrdinalElection, election_2: OrdinalElection,
                         distance_id: str = None, **kwargs) -> float or (float, list):
    """
    Computes distance between ordinal elections, (if applicable) optimal matching.

    :param election_1: first election.
    :param election_2: second election.
    :param distance_id: name of the distance.
    :param kwargs: additional arguments.
    :return: distances, matching (if applicable).
    """

    inner_distance, main_distance = _extract_distance_id(distance_id)

    if main_distance in registered_ordinal_distances:

        if inner_distance is not None:
            return registered_ordinal_distances.get(main_distance)(election_1,
                                                                   election_2,
                                                                   inner_distance)
        else:
            return registered_ordinal_distances.get(main_distance)(election_1,
                                                                   election_2,
                                                                   **kwargs)

    else:
        logging.warning(f'No such distance as: {main_distance}!')


def _extract_distance_id(distance_id: str) -> (Callable, str):
    """ Return: inner distance (distance between votes) name and main distance name """
    if '-' in distance_id:
        inner_distance, main_distance = distance_id.split('-')
        inner_distance = map_str_to_func(inner_distance)
    else:
        main_distance = distance_id
        inner_distance = None
    return inner_distance, main_distance


def run_single_process(exp: Experiment,
                       instances_ids: list,
                       distances: dict,
                       times: dict,
                       matchings: dict,
                       safe_mode=False) -> None:
    """ Single process for computing distances """

    for instance_id_1, instance_id_2 in tqdm(instances_ids, desc='Computing distances'):
        start_time = time()
        if safe_mode:
            distance = get_distance(copy.deepcopy(exp.instances[instance_id_1]),
                                    copy.deepcopy(exp.instances[instance_id_2]),
                                    distance_id=copy.deepcopy(exp.distance_id))
        else:
            distance = get_distance(exp.instances[instance_id_1],
                                    exp.instances[instance_id_2],
                                    distance_id=exp.distance_id)
        if type(distance) is tuple:
            distance, matching = distance
            matching = np.array(matching)
            matchings[instance_id_1][instance_id_2] = matching
            matchings[instance_id_2][instance_id_1] = np.argsort(matching)
        distances[instance_id_1][instance_id_2] = distance
        distances[instance_id_2][instance_id_1] = distances[instance_id_1][instance_id_2]
        times[instance_id_1][instance_id_2] = time() - start_time
        times[instance_id_2][instance_id_1] = times[instance_id_1][instance_id_2]


def run_multiple_processes(experiment: Experiment,
                           instances_ids: list,
                           distances: dict,
                           times: dict,
                           matchings: dict,
                           process_id) -> None:
    """ Single process for computing distances """

    for instance_id_1, instance_id_2 in tqdm(instances_ids,
                                             desc=f'Computing distances of thread {process_id}'):
        start_time = time()
        distance = get_distance(copy.deepcopy(experiment.instances[instance_id_1]),
                                copy.deepcopy(experiment.instances[instance_id_2]),
                                distance_id=copy.deepcopy(experiment.distance_id))
        if type(distance) is tuple:
            distance, matching = distance
            matching = np.array(matching)
            matchings[instance_id_1][instance_id_2] = matching
            matchings[instance_id_2][instance_id_1] = np.argsort(matching)
        distances[instance_id_1][instance_id_2] = distance
        distances[instance_id_2][instance_id_1] = distances[instance_id_1][instance_id_2]
        times[instance_id_1][instance_id_2] = time() - start_time
        times[instance_id_2][instance_id_1] = times[instance_id_1][instance_id_2]

    if experiment.is_exported:
        exports.export_distances_multiple_processes(experiment,
                                                    instances_ids,
                                                    distances,
                                                    times,
                                                    process_id)



import sys
import time
import os

try:
    from dotenv import load_dotenv
    load_dotenv()
    sys.path.append(os.environ["PATH"])
    from abcvoting import fileio, genprofiles
    import gurobipy as gb
    from abcvoting.preferences import Profile, Voter
    from abcvoting import abcrules
except ImportError:
    pass


def convert_election_to_profile(election):
    profile = Profile(num_cand=election.num_candidates)

    voters = []
    for i, vote in enumerate(election.votes):
        voter = Voter(vote)
        voters.append(voter)

    profile._voters = voters
    return profile


def partylistdistance(election, feature_params=None):

    if 'largepartysize' in feature_params:
        largepartysize = feature_params['largepartysize']
    else:
        largepartysize = 5

    if 'time_limit' in feature_params:
        time_limit = feature_params['time_limit']
    else:
        time_limit = 5

    profile = convert_election_to_profile(election)

    model = gb.Model()
    model.setParam("OutputFlag", False)
    model.setParam('TimeLimit', time_limit)  # in seconds

    same_party = {}
    for c1 in profile.candidates:
        for c2 in range(c1):
            same_party[(c1, c2)] = model.addVar(vtype=gb.GRB.BINARY)

    edit = {}
    for v, _ in enumerate(profile):
        for c in profile.candidates:
            edit[(v, c)] = model.addVar(vtype=gb.GRB.BINARY)

    for v, vote in enumerate(profile):
        for c1 in profile.candidates:
            for c2 in range(c1):
                if c1 in vote.approved and c2 in vote.approved:
                    model.addConstr(
                        (same_party[(c1, c2)] == 1) >> (edit[(v, c1)] == edit[(v, c2)])
                    )
                    model.addConstr(
                        # (same_party[(c1, c2)] == 0) >> (edit[(v, c1)] + edit[(v, c2)] >= 1)
                        (same_party[(c1, c2)] + edit[(v, c1)] + edit[(v, c2)] >= 1)
                    )
                elif c1 not in vote.approved and c2 not in vote.approved:
                    model.addConstr(
                        (same_party[(c1, c2)] == 1) >> (edit[(v, c1)] == edit[(v, c2)])
                    )
                    model.addConstr(
                        # (same_party[(c1, c2)] == 0) >> (edit[(v, c1)] + edit[(v, c2)] <= 1)
                        (edit[(v, c1)] + edit[(v, c2)] <= 1 + same_party[(c1, c2)])
                    )
                else:
                    model.addConstr(
                        (same_party[(c1, c2)] == 1) >> (edit[(v, c1)] + edit[(v, c2)] == 1)
                    )
                    if c1 in vote.approved:
                        model.addConstr(
                            (same_party[(c1, c2)] + edit[(v, c1)] >= edit[(v, c2)])
                            # (same_party[(c1, c2)] == 0) >> (edit[(v, c1)] >= edit[(v, c2)])
                        )
                    else:
                        model.addConstr(
                            (same_party[(c1, c2)] + edit[(v, c2)] >= edit[(v, c1)])
                            # (same_party[(c1, c2)] == 0) >> (edit[(v, c2)] >= edit[(v, c1)])
                        )

    model.setObjective(
        gb.quicksum(edit[(v, c)] for v, _ in enumerate(profile) for c in profile.candidates),
        gb.GRB.MINIMIZE,
    )

    model.optimize()

    newprofile = Profile(num_cand=profile.num_cand)
    for v, vote in enumerate(profile):
        new_approved = {
            c
            for c in profile.candidates
            if (c in vote.approved and edit[(v, c)].X <= 0.1)
               or (c not in vote.approved and edit[(v, c)].X >= 0.9)
        }
        newprofile.add_voter(new_approved)

    # number of parties
    parties = set(profile.candidates)
    for c1 in profile.candidates:
        for c2 in range(c1):
            if same_party[(c1, c2)].X >= 0.9:
                parties.discard(c2)
    num_large_parties = 0
    for party in parties:
        support = len([voter for voter in newprofile if party in voter.approved])
        if support >= largepartysize:
            num_large_parties += 1


    print(model.objVal, model.objbound, num_large_parties)
    return model.objVal, model.objbound, num_large_parties


def pav_time(election):

    profile = Profile(election.num_candidates)
    profile.add_voters(election.votes)
    committee_size = 10
    resolute = True
    rule_name = 'pav'
    start = time.time()
    winning_committees = abcrules.compute(rule_name, profile, committee_size,
                                          algorithm="gurobi", resolute=resolute)
    return time.time() - start



import csv
import logging
from collections import Counter

from matplotlib import pyplot as plt

import mapel.elections.persistence.election_exports as exports
import mapel.elections.persistence.election_imports as imports
from mapel.core.inner_distances import swap_distance_between_potes, \
    spearman_distance_between_potes
from mapel.core.utils import *
from mapel.elections.cultures.fake import *
from mapel.elections.cultures.matrices.group_separable_matrices import get_gs_caterpillar_vectors
from mapel.elections.cultures.preflib import get_sushi_vectors
from mapel.elections.cultures.matrices.single_crossing_matrices import get_single_crossing_vectors
from mapel.elections.cultures_ import generate_ordinal_votes, \
    from_approval, generate_ordinal_alliance_votes
from mapel.elections.features.other import is_condorcet
from mapel.elections.objects.Election import Election
from mapel.elections.other.winners import compute_sntv_winners, compute_borda_winners, \
    compute_stv_winners
from mapel.elections.other.winners import generate_winners
from mapel.elections.other.winners import get_borda_points


class OrdinalElection(Election):

    def __init__(self,
                 experiment_id=None,
                 election_id=None,
                 culture_id=None,
                 votes=None,
                 label=None,
                 num_voters: int = None,
                 num_candidates: int = None,
                 variable=None,
                 fast_import=False,
                 **kwargs):

        super().__init__(experiment_id=experiment_id,
                         election_id=election_id,
                         culture_id=culture_id,
                         votes=votes,
                         label=label,
                         num_voters=num_voters,
                         num_candidates=num_candidates,
                         fast_import=fast_import,
                         **kwargs)

        self.variable = variable
        self.vectors = []
        self.matrix = []
        self.potes = None
        self.condorcet = None
        self.points = {}
        self.alliances = {}

        self.import_ordinal_election()

    def import_ordinal_election(self):

        if self.is_imported and self.experiment_id != 'virtual':

            try:
                if self.votes is not None:
                    self.culture_id = self.culture_id
                    if str(self.votes[0]) in LIST_OF_FAKE_MODELS:
                        self.fake = True
                        self.votes = self.votes[0]
                        self.num_candidates = self.votes[1]
                        self.num_voters = self.votes[2]
                    else:
                        self.votes = self.votes
                        self.num_candidates = len(self.votes[0])
                        self.num_voters = len(self.votes)
                        self.compute_potes()
                else:

                    try:
                        self.fake = imports.check_if_fake(self.experiment_id, self.election_id,
                                                          'soc')
                    except:
                        self.fake = False

                    if self.fake:
                        self.culture_id, self.params, self.num_voters, \
                        self.num_candidates = imports.import_fake_soc_election(self.experiment_id,
                                                                               self.election_id)
                    else:
                        self.votes, self.num_voters, self.num_candidates, self.params, \
                        self.culture_id, self.alliances, \
                        self.num_options, self.quantites, \
                        self.distinct_votes = imports.import_real_soc_election(
                            experiment_id=self.experiment_id,
                            election_id=self.election_id,
                            is_shifted=self.is_shifted)
                        try:
                            self.points['voters'] = self.import_ideal_points('voters')
                            self.points['candidates'] = self.import_ideal_points('candidates')
                        except:
                            pass

                        try:
                            self.alpha = 1
                            if self.params and 'alpha' in self.params:
                                self.alpha = self.params['alpha']
                        except KeyError:
                            print("Error")

                self.candidatelikeness_original_vectors = {}

                if not self.fast_import:
                    self.votes_to_positionwise_vectors()
            except:
                self.is_correct = False

        if self.params is None:
            self.params = {}

        try:
            self.params, self.printing_params = update_params_ordinal(
                self.params,
                self.printing_params,
                self.variable,
                self.culture_id,
                self.num_candidates)
        except:
            pass

    def get_vectors(self):
        if self.vectors is not None and len(self.vectors) > 0:
            return self.vectors
        return self.votes_to_positionwise_vectors()

    def get_matrix(self):
        if self.matrix is not None and len(self.matrix) > 0:
            return self.matrix
        return self.votes_to_positionwise_matrix()

    def get_potes(self):
        if self.potes is not None:
            return self.potes
        return self.compute_potes()

    def votes_to_positionwise_vectors(self):
        vectors = np.zeros([self.num_candidates, self.num_candidates])

        if self.culture_id == 'conitzer_matrix':
            vectors = get_conitzer_vectors(self.num_candidates)
        elif self.culture_id == 'walsh_matrix':
            vectors = get_walsh_vectors(self.num_candidates)
        elif self.culture_id == 'single-crossing_matrix':
            vectors = get_single_crossing_vectors(self.num_candidates)
        elif self.culture_id == 'gs_caterpillar_matrix':
            vectors = get_gs_caterpillar_vectors(self.num_candidates)
        elif self.culture_id == 'sushi_matrix':
            vectors = get_sushi_vectors()
        elif self.culture_id in {'norm-mallows_matrix', 'mallows_matrix_path'}:
            vectors = get_mallows_vectors(self.num_candidates, self.params)
        elif self.culture_id in {'identity', 'uniformity', 'antagonism', 'stratification'}:
            vectors = get_fake_vectors_single(self.culture_id, self.num_candidates)
        elif self.culture_id in {'walsh_path', 'conitzer_path'}:
            vectors = get_fake_multiplication(self.num_candidates, self.params,
                                              self.culture_id)
        elif self.culture_id in PATHS:
            vectors = get_fake_convex(self.culture_id, self.num_candidates, self.num_voters,
                                      self.params, get_fake_vectors_single)
        elif self.culture_id == 'crate':
            vectors = get_fake_vectors_crate(num_candidates=self.num_candidates,
                                             fake_param=self.params)
        elif self.culture_id in ['from_approval']:
            vectors = from_approval(num_candidates=self.num_candidates,
                                    num_voters=self.num_voters,
                                    params=self.params)
        else:
            for i in range(self.num_voters):
                pos = 0
                for j in range(self.num_candidates):
                    vote = self.votes[i][j]
                    if vote == -1:
                        continue
                    vectors[vote][pos] += 1
                    pos += 1
            for i in range(self.num_candidates):
                for j in range(self.num_candidates):
                    vectors[i][j] /= float(self.num_voters)

        self.vectors = vectors
        self.matrix = self.vectors.transpose()
        return vectors

    def votes_to_positionwise_matrix(self):
        return self.votes_to_positionwise_vectors().transpose()

    def votes_to_pairwise_matrix(self) -> np.ndarray:
        """ convert VOTES to pairwise MATRIX """
        matrix = np.zeros([self.num_candidates, self.num_candidates])
        if self.fake:
            if self.culture_id in {'identity', 'uniformity', 'antagonism', 'stratification'}:
                matrix = get_fake_matrix_single(self.culture_id, self.num_candidates)
            elif self.culture_id in PATHS:
                matrix = get_fake_convex(self.culture_id,
                                         self.num_candidates,
                                         self.num_voters,
                                         self.fake_param,
                                         get_fake_matrix_single)

        else:
            for v in range(self.num_voters):
                for c1 in range(self.num_candidates):
                    for c2 in range(c1 + 1, self.num_candidates):
                        matrix[int(self.votes[v][c1])][
                            int(self.votes[v][c2])] += 1
            for i in range(self.num_candidates):
                for j in range(i + 1, self.num_candidates):
                    matrix[i][j] /= float(self.num_voters)
                    matrix[j][i] = 1. - matrix[i][j]
        return matrix

    def votes_to_bordawise_vector(self) -> np.ndarray:
        """ convert VOTES to Borda vector """
        borda_vector = np.zeros([self.num_candidates])
        if self.fake:
            if self.culture_id in {'identity', 'uniformity', 'antagonism', 'stratification'}:
                borda_vector = get_fake_borda_vector(self.culture_id, self.num_candidates,
                                                     self.num_voters)
            elif self.culture_id in PATHS:
                borda_vector = get_fake_convex(self.culture_id,
                                               self.num_candidates,
                                               self.num_voters,
                                               self.params,
                                               get_fake_borda_vector)
        else:
            c = self.num_candidates
            v = self.num_voters
            vectors = self.votes_to_positionwise_vectors()
            borda_vector = [sum([vectors[i][j] * (c - j - 1) for j in range(c)]) * v for i in
                            range(self.num_candidates)]
            borda_vector = sorted(borda_vector, reverse=True)

        return np.array(borda_vector)

    def votes_to_candidatelikeness_original_vectors(self) -> None:
        """ convert VOTES to candidate-likeness VECTORS """
        matrix = np.zeros([self.num_candidates, self.num_candidates])

        for c_1 in range(self.num_candidates):
            for c_2 in range(self.num_candidates):
                for vote in self.approval_votes:
                    if (c_1 in vote and c_2 not in vote) or (c_1 not in vote and c_2 in vote):
                        matrix[c_1][c_2] += 1
        self.candidatelikeness_original_vectors = matrix / self.num_voters

    def votes_to_positionwise_intervals(self, precision: int = None) -> list:

        vectors = self.votes_to_positionwise_matrix()
        return [self.vector_to_interval(vectors[i], precision=precision)
                for i in range(len(vectors))]

    def votes_to_voterlikeness_vectors(self) -> np.ndarray:
        return self.votes_to_voterlikeness_matrix()

    def votes_to_voterlikeness_matrix(self, vote_distance='swap') -> np.ndarray:
        """ convert VOTES to voter-likeness MATRIX """
        matrix = np.zeros([self.num_voters, self.num_voters])
        self.compute_potes()

        for v1 in range(self.num_voters):
            for v2 in range(self.num_voters):
                if vote_distance == 'swap':
                    matrix[v1][v2] = swap_distance_between_potes(self.potes[v1], self.potes[v2])
                elif vote_distance == 'spearman':
                    matrix[v1][v2] = spearman_distance_between_potes(self.potes[v1], self.potes[v2])

        for i in range(self.num_voters):
            for j in range(i + 1, self.num_voters):
                matrix[j][i] = matrix[i][j]

        return matrix

    def votes_to_agg_voterlikeness_vector(self):
        """ convert VOTES to Borda vector """

        vector = np.zeros([self.num_voters])

        for v1 in range(self.num_voters):
            for v2 in range(self.num_voters):

                swap_distance = 0
                for i in range(self.num_candidates):
                    for j in range(i + 1, self.num_candidates):
                        if (self.potes[v1][i] > self.potes[v1][j] and
                            self.potes[v2][i] < self.potes[v2][j]) or \
                                (self.potes[v1][i] < self.potes[v1][j] and
                                 self.potes[v2][i] > self.potes[v2][j]):
                            swap_distance += 1
                vector[v1] += swap_distance

        return vector, len(vector)

    def compute_winners(self, method=None, num_winners=None):

        self.borda_points = get_borda_points(self.votes, self.num_voters, self.num_candidates)

        if method == 'sntv':
            self.winners = compute_sntv_winners(election=self, num_winners=num_winners)
        if method == 'borda':
            self.winners = compute_borda_winners(election=self, num_winners=num_winners)
        if method == 'stv':
            self.winners = compute_stv_winners(election=self, num_winners=num_winners)
        if method in {'approx_cc', 'approx_hb', 'approx_pav'}:
            self.winners = generate_winners(election=self, num_winners=num_winners)

    def prepare_instance(self, is_exported=None, is_aggregated=True):
        if 'num_alliances' in self.params:
            self.votes, self.alliances = generate_ordinal_alliance_votes(
                culture_id=self.culture_id,
                num_candidates=self.num_candidates,
                num_voters=self.num_voters,
                params=self.params)
        else:
            self.votes = generate_ordinal_votes(culture_id=self.culture_id,
                                                num_candidates=self.num_candidates,
                                                num_voters=self.num_voters,
                                                params=self.params)
        if not self.fake:
            c = Counter(map(tuple, self.votes))
            counted_votes = [[count, list(row)] for row, count in c.items()]
            counted_votes = sorted(counted_votes, reverse=True)
            self.quantites = [a[0] for a in counted_votes]
            self.distinct_votes = [a[1] for a in counted_votes]
            self.num_options = len(counted_votes)
        else:
            self.quantites = [self.num_voters]
            self.num_options = 1

        if is_exported:
            exports.export_ordinal_election(self, is_aggregated=is_aggregated)

    def compute_distances(self, distance_id='swap', object_type=None):
        """ Return: distances between votes """
        if object_type is None:
            object_type = self.object_type

        self.distinct_potes = convert_votes_to_potes(self.distinct_votes)
        self.num_dist_votes = len(self.distinct_votes)
        self.num_options = self.num_dist_votes

        if object_type == 'vote':
            distances = np.zeros([self.num_dist_votes, self.num_dist_votes])
            for v1 in range(self.num_dist_votes):
                for v2 in range(self.num_dist_votes):
                    if distance_id == 'swap':
                        distances[v1][v2] = swap_distance_between_potes(
                            self.distinct_potes[v1], self.distinct_potes[v2])
                    elif distance_id == 'spearman':
                        distances[v1][v2] = spearman_distance_between_potes(
                            self.distinct_potes[v1], self.distinct_potes[v2])
        elif object_type == 'candidate':
            self.compute_potes()
            if distance_id == 'domination':
                distances = self.votes_to_pairwise_matrix()
                distances = np.abs(distances - 0.5) * self.num_voters
                np.fill_diagonal(distances, 0)
            elif distance_id == 'position':
                distances = np.zeros([self.num_candidates, self.num_candidates])
                for c1 in range(self.num_candidates):
                    for c2 in range(self.num_candidates):
                        dist = 0
                        for pote in self.potes:
                            dist += abs(pote[c1] - pote[c2])
                        distances[c1][c2] = dist
        else:
            logging.warning('incorrect object_type')

        self.distances[object_type] = distances

        if object_type == 'vote':
            length = self.num_dist_votes
        elif object_type == 'candidate':
            length = self.num_candidates

        if self.is_exported:
            exports.export_distances(self, object_type=object_type, length=length)

    def is_condorcet(self):
        """ Check if election witness Condorcet winner"""
        if self.condorcet is None:
            self.condorcet = is_condorcet(self)
        return self.condorcet

    def import_ideal_points(self, name):
        path = os.path.join(os.getcwd(), "experiments", self.experiment_id, "elections",
                            f'{self.election_id}_{name}.csv')
        points = []
        with open(path, 'r', newline='') as csv_file:
            reader = csv.DictReader(csv_file, delimiter=';')
            for row in reader:
                points.append([float(row['x']), float(row['y'])])
        return points


    @staticmethod
    def texify_label(name):
        return name.replace('phi', '$\phi$'). \
            replace('alpha', '$\\ \\alpha$'). \
            replace('omega', '$\\ \\omega$'). \
            replace('§', '\n', 1). \
            replace('0.005', '$\\frac{1}{200}$'). \
            replace('0.025', '$\\frac{1}{40}$'). \
            replace('0.75', '$\\frac{3}{4}$'). \
            replace('0.25', '$\\frac{1}{4}$'). \
            replace('0.01', '$\\frac{1}{100}$'). \
            replace('0.05', '$\\frac{1}{20}$'). \
            replace('0.5', '$\\frac{1}{2}$'). \
            replace('0.1', '$\\frac{1}{10}$'). \
            replace('0.2', '$\\frac{1}{5}$'). \
            replace('0.4', '$\\frac{2}{5}$'). \
            replace('0.8', '$\\frac{4}{5}$'). \
            replace(' ', '\n', 1)

    def print_map(
            self,
            show=True,
            radius=None,
            name=None,
            alpha=0.1,
            s=30,
            circles=False,
            object_type=None,
            double_gradient=False,
            saveas=None,
            color='blue',
            marker='o',
            title_size=20
    ):

        if object_type is None:
            object_type = self.object_type

        if object_type == 'vote':
            length = self.num_voters
        elif object_type == 'candidate':
            length = self.num_candidates
        else:
            logging.warning(f'Incorrect object type: {object_type}')
            length = 0

        plt.figure(figsize=(6.4, 6.4))

        X = []
        Y = []
        for elem in self.coordinates[object_type]:
            X.append(elem[0])
            Y.append(elem[1])

        start = False
        if start:
            plt.scatter(X[0], Y[0],
                        color='sienna',
                        s=1000,
                        alpha=1,
                        marker='X')

        if object_type == 'vote':
            if double_gradient:
                for i in range(length):
                    x = float(self.points['voters'][i][0])
                    y = float(self.points['voters'][i][1])
                    plt.scatter(X[i], Y[i], color=[0, y, x], s=s, alpha=alpha)
            else:
                for i in range(len(X)):
                    plt.scatter(X[i], Y[i], color=color, alpha=alpha, marker=marker,
                                s=self.quantites[i] * s)

        elif object_type == 'candidate':
            for i in range(len(X)):
                plt.scatter(X[i], Y[i], color=color, alpha=alpha, marker=marker,
                            s=s)

        avg_x = np.mean(X)
        avg_y = np.mean(Y)

        if radius:
            plt.xlim([avg_x - radius, avg_x + radius])
            plt.ylim([avg_y - radius, avg_y + radius])
        # plt.title(election.label, size=38)

        try:
            plt.title(self.texify_label(self.label), size=title_size)  # tmp
        except:
            pass

        # plt.title(election.texify_label(election.label), size=38, y=0.94)
        # plt.title(election.label, size=title_size)
        plt.axis('off')

        if saveas is not None:

            if saveas == 'default':

                dir = os.path.join(os.getcwd(), "images", self.experiment_id)
                if not os.path.isdir(dir):
                    os.mkdir(os.path.join(os.getcwd(), dir))

                saveas = f'{self.label}_{object_type}'
                file_name = os.path.join(os.getcwd(), "images", self.experiment_id, f'{saveas}.png')
                plt.savefig(file_name, bbox_inches='tight', dpi=100)

        if show:
            plt.show()
        else:
            plt.clf()

        plt.close()



def convert_votes_to_potes(votes) -> np.array:
    """ Convert votes to positional votes (called potes) """
    return np.array([[list(vote).index(i) for i, _ in enumerate(vote)]
                     for vote in votes])



import math
import numpy as np
import mapel.elections.cultures.mallows as mallows
import itertools as it
import mapel.elections.cultures.sampling.samplemat as smpl


# Auxiliary functions

def distribute_in_matrix(n, m):
    if m == 0:
        return []
    k = n // m
    r = n - k * m
    matrix = []
    for i in range(m):
        row = [k for _ in range(m)]
        for j in range(i, i + r):
            if j >= m:
                j = j - m
            row[j] = row[j] + 1
        matrix.append(row)
    return matrix


def distribute_in_block_matrix(n, blocks):
    before = 0
    after = sum(blocks)
    matrix = []
    for b in blocks:
        after = after - b
        block = distribute_in_matrix(n, b)
        for row in block:
            matrix.append([0 for _ in range(before)] + row + [0 for _ in range(after)])
        before = before + b
    return (matrix)


def draw_election(matrix):
    return smpl.sample_election_using_permanent(matrix)


def build_perms(matrix):
    if len(matrix[0]) == 0:
        return [[]]
    perms = []
    last_col = []
    for j in range(len(matrix)):
        last_col.append(matrix[j].pop())
    if sum(last_col) > 0:
        for last_cand, elig in enumerate(last_col):
            if elig <= 0:
                continue
            zeroed_vector = matrix[last_cand][:]
            matrix[last_cand][:] = len(matrix[last_cand]) * [0]
            for perm in build_perms(matrix):
                perms.append(perm + [last_cand])
            matrix[last_cand][:] = zeroed_vector[:]
    for j in range(len(matrix)):
        matrix[j].append(last_col[j])
    return perms


# Models
def generate_un_from_list(num_voters=None, num_candidates=None):
    id_perm = list(range(num_candidates))
    m_fac = math.factorial(num_candidates)
    alls = num_voters // m_fac
    rest = num_voters - alls * m_fac
    res = []
    for _ in range(alls):
        res = res + [list(v) for v in it.permutations(id_perm)]
    res = res + [list(v) for v in it.permutations(id_perm)][:rest]
    return res


def generate_un_from_matrix_votes(num_voters=None, num_candidates=None):
    """ Generate real elections that have UN positionwise matrix """
    matrix = distribute_in_matrix(num_voters, num_candidates)
    return draw_election(matrix)


def generate_idan_part_votes(num_voters=None, num_candidates=None, part_share=None, **kwargs):
    """ Generate real elections between (ID) and (AN) """
    if part_share is None:
        print("IDAN_part generation : params None : random param generated")
        part_size = np.random.choice(range(num_voters))
    else:
        part_size = part_share * (num_voters)
    part_size = int(round(part_size))
    id_share = num_voters - (part_size // 2)
    op_share = part_size // 2
    votes = [[j for j in range(num_candidates)] for _ in range(id_share)]
    votes = votes + [[(num_candidates - j - 1) for j in range(num_candidates)] for _ in
                     range(op_share)]
    return votes


def generate_idun_part_votes(num_voters=None, num_candidates=None, part_share=None, **kwargs):
    """ Generate elections realizing linear combinations of pos-matrices between (ID) and (UN) """
    if part_share is None:
        print("IDUN_part generation : params None : random param generated")
        part_size = np.random.choice(range(num_voters))
    else:
        part_size = part_share * (num_voters)
    part_size = int(round(part_size))
    id_share = num_voters - part_size
    un_share = part_size
    votes = [[j for j in range(num_candidates)] for _ in range(id_share)]
    votes = votes + draw_election(distribute_in_matrix(un_share, num_candidates))
    return votes


def generate_idst_part_votes(num_voters=None, num_candidates=None, part_share=None, **kwargs):
    """ Generate elections realizing linear combinations of pos-matrices between (ID) and (ST) """
    if part_share is None:
        print("IDST_part generation : params None : random param generated")
        part_size = np.random.choice(range(num_voters))
    else:
        part_size = part_share * (num_voters)
    part_size = int(round(part_size))
    id_share = num_voters - part_size
    st_share = part_size
    topsize = num_candidates // 2
    bottomsize = num_candidates - topsize
    votes_id = [[j for j in range(num_candidates)] for _ in range(id_share)]
    votes_st = draw_election(distribute_in_block_matrix(st_share, [topsize, bottomsize]))
    return votes_id + votes_st


def generate_anun_part_votes(num_voters=None, num_candidates=None, part_share=None, **kwargs):
    """ Generate elections realizing linear combinations of pos-matrices between (AN) and (UN) """
    if part_share is None:
        print("ANUN_part generation : params None : random param generated")
        part_size = np.random.choice(range(num_voters))
    else:
        part_size = part_share * (num_voters)
    part_size = int(round(part_size))
    id_share = (num_voters - part_size) // 2
    op_share = num_voters - part_size - id_share
    un_share = num_voters - id_share - op_share
    votes = [[j for j in range(num_candidates)] for _ in range(id_share)]
    votes = votes + [[(num_candidates - j - 1) for j in range(num_candidates)] for _ in
                     range(op_share)]
    votes = votes + draw_election(distribute_in_matrix(un_share, num_candidates))
    return votes


def generate_anst_part_votes(num_voters=None, num_candidates=None, part_share=None, **kwargs):
    """ Generate elections realizing linear combinations of pos-matrices between (AN) and (ST) """
    if part_share is None:
        print("ANST_part generation : params None : random param generated")
        part_size = np.random.choice(range(num_voters))
    else:
        part_size = part_share * (num_voters)
    part_size = int(round(part_size))
    id_share = (num_voters - part_size) // 2
    op_share = num_voters - part_size - id_share
    st_share = num_voters - id_share - op_share
    topsize = num_candidates // 2
    bottomsize = num_candidates - topsize
    votes = [[j for j in range(num_candidates)] for _ in range(id_share)]
    votes = votes + [[(num_candidates - j - 1) for j in range(num_candidates)] for _ in
                     range(op_share)]
    votes = votes + draw_election(distribute_in_block_matrix(st_share, [topsize, bottomsize]))
    return votes


def generate_unst_part_votes(num_voters=None, num_candidates=None, part_share=None, **kwargs):
    """ Generate elections realizing linear combinations of pos-matrices between (UN) and (ST) """
    if part_share is None:
        print("UNST_part generation : params None : random param generated")
        part_size = np.random.choice(range(num_voters))
    else:
        part_size = part_share * (num_voters)
    part_size = int(round(part_size))
    un_share = num_voters - part_size
    st_share = part_size
    topsize = num_candidates // 2
    bottomsize = num_candidates - topsize
    votes = draw_election(distribute_in_matrix(un_share, num_candidates))
    votes = votes + draw_election(distribute_in_block_matrix(st_share, [topsize, bottomsize]))
    return votes


def generate_idan_mallows_votes(num_voters=None, num_candidates=None, params=None):
    if params is None or not ('scaled-phi' in params):
        print("IDAN_mallows generation : params None : random param generated")
        is_reversed = np.random.choice([True, False])
        phi = mallows.phi_from_relphi(num_candidates, relphi=np.random.uniform())
    else:
        if params['scaled-phi'] > 0.5:
            is_reversed = True
            phi = mallows.phi_from_relphi(num_candidates, (1 - params['scaled-phi']) * 2)
        else:
            is_reversed = False
            phi = mallows.phi_from_relphi(num_candidates, params['scaled-phi'] * 2)
    id_share = num_voters // 2
    op_share = num_voters - id_share
    votes_id = [[j for j in range(num_candidates)] for _ in range(id_share)]
    votes_op = mallows.generate_mallows_votes(op_share, num_candidates, {'phi': phi})
    if is_reversed:
        for v in votes_op:
            v.reverse()
    return votes_id + votes_op


def generate_idst_mallows_votes(num_voters=None, num_candidates=None, params=None):
    if params is None or not ('phi' in params):
        print("IDST_mallows generation : params None : random param generated")
        phi = mallows.phi_from_relphi(num_candidates, relphi=np.random.uniform())
    else:
        phi = params['phi']
    better = num_candidates // 2
    worse = num_candidates - better
    votes_better = mallows.generate_mallows_votes(num_voters, better, {'phi': phi})
    votes_worse = mallows.generate_mallows_votes(num_voters, worse, {'phi': phi})
    votes = []
    for b, w in zip(votes_better, votes_worse):
        w_ = [c + better for c in w]
        v = b + w_
        votes.append(v)
    return votes


def generate_anun_mallows_votes(num_voters=None, num_candidates=None, params=None):
    mallows_params = {}
    if params is None or not ('phi' in params):
        print("IDST_mallows generation : params None : random param generated")
        mallows_params['phi'] = mallows.phi_from_relphi(num_candidates, relphi=np.random.uniform())
    else:
        mallows_params['phi'] = params['phi']
    mallows_params['weight'] = 0.5
    return mallows.generate_mallows_votes(num_voters, num_candidates, mallows_params)


def generate_unst_mallows_votes(num_voters=None, num_candidates=None, params=None):
    if params is None or not ('phi' in params):
        print("IDST_mallows generation : params None : random param generated")
        phi = mallows.phi_from_relphi(num_candidates, relphi=np.random.uniform()) / 2
    else:
        phi = params['phi'] / 2
    better = num_candidates // 2
    worse = num_candidates - better
    votes = draw_election(distribute_in_block_matrix(num_voters, [better, worse]))
    # The next part works poorly for odd number of candidates (the last one is always from worse part)
    for v in votes:
        for i in range(better):
            if np.random.random() < phi:
                c = v[i]
                v[i] = v[i + better]
                v[i + better] = c
    return votes


def generate_unst_topsize_votes(num_voters=None, num_candidates=None, params=None):
    """ Generate kind of real elections between (UN) and (ST) """
    if params is None or not ('top_share' in params):
        print("UNST_topsize generation : params None : random param generated")
        top_share = np.random.random()
    else:
        top_share = params['top_share']
    top_size = int(round(top_share * num_candidates))
    better = top_size
    worse = num_candidates - top_size
    matrix = distribute_in_block_matrix(num_voters, [better, worse])
    return draw_election(matrix)


def generate_idst_blocks_votes(num_voters=None, num_candidates=None, params=None):
    """ Generate kind of real elections between (ID) and (UN) """
    if params is None or not ('no_blocks' in params):
        print("IDST_blocks generation : params None : random param generated")
        no_blocks = np.random.choice(range(num_candidates + 1))
    else:
        no_blocks = params['no_blocks']
    no_blocks = int(round(no_blocks))
    k = num_candidates // no_blocks
    r = num_candidates - k * no_blocks
    blocks = [k for _ in range(no_blocks)]
    with_one_more = list(np.random.choice(range(no_blocks), r, replace=False))
    for i in with_one_more:
        blocks[i] = blocks[i] + 1
    print(blocks)
    matrix = distribute_in_block_matrix(num_voters, blocks)
    return draw_election(matrix)



#!/usr/bin/env python

import mapel.elections.features.approx as approx
import mapel.elections.features.banzhaf_cc as banzhaf_cc
import mapel.elections.features.clustering as clustering
import mapel.elections.features.cohesive as cohesive
import mapel.elections.features.dimensionality as dimensionality
import mapel.elections.features.distortion as distortion
import mapel.elections.features.diversity as diversity
import mapel.elections.features.justified_representation as jr
import mapel.elections.features.other as other
import mapel.elections.features.partylist as partylist
import mapel.elections.features.proportionality_degree as prop_deg
import mapel.elections.features.ranging_cc as ranging_cc
import mapel.elections.features.scores as scores
import mapel.elections.features.vc_diversity as vcd
from mapel.core.features_main import get_main_local_feature, get_main_global_feature
from mapel.core.glossary import MAIN_LOCAL_FEATUERS, MAIN_GLOBAL_FEATUERS

registered_approval_features = {
    'max_approval_score': other.max_approval_score,
    'number_of_cohesive_groups': cohesive.count_number_of_cohesive_groups,
    'number_of_cohesive_groups_brute': cohesive.count_number_of_cohesive_groups_brute,
    'proportionality_degree_av': prop_deg.proportionality_degree_av,
    'proportionality_degree_pav': prop_deg.proportionality_degree_pav,
    'proportionality_degree_cc': prop_deg.proportionality_degree_cc,

    'cohesiveness': cohesive.count_largest_cohesiveness_level_l_of_cohesive_group,
    # unsupported feature
    'ejr': jr.test_ejr,  # unsupported feature
}

registered_ordinal_features = {
    'highest_borda_score': scores.highest_borda_score,
    'highest_plurality_score': scores.highest_plurality_score,
    'highest_copeland_score': scores.highest_copeland_score,
    'lowest_dodgson_score': scores.lowest_dodgson_score,
    'highest_cc_score': scores.highest_cc_score,
    'highest_hb_score': scores.highest_hb_score,
    'highest_pav_score': scores.highest_pav_score,
    'greedy_approx_cc_score': approx.get_greedy_approx_cc_score,
    'removal_approx_cc_score': approx.get_removal_approx_cc_score,
    'greedy_approx_hb_score': approx.get_greedy_approx_hb_score,
    'removal_approx_hb_score': approx.get_removal_approx_hb_score,
    'greedy_approx_pav_score': approx.get_greedy_approx_pav_score,
    'removal_approx_pav_score': approx.get_removal_approx_pav_score,
    'banzhaf_cc_score': banzhaf_cc.get_banzhaf_cc_score,
    'ranging_cc_score': ranging_cc.get_ranging_cc_score,
    'num_of_diff_votes': vcd.num_of_diff_votes,
    'voterlikeness_sqrt': vcd.voterlikeness_sqrt,
    'voterlikeness_harmonic': vcd.voterlikeness_harmonic,
    'borda_diversity': vcd.borda_diversity,
    'borda_std': diversity.borda_std,
    'borda_range': diversity.borda_range,
    'borda_gini': diversity.borda_gini,
    'borda_meandev': diversity.borda_meandev,
    'cand_dom_dist_mean': diversity.cand_dom_dist_mean,
    'cand_dom_dist_std': diversity.cand_dom_dist_std,
    'cand_pos_dist_std': diversity.cand_pos_dist_std,
    'cand_pos_dist_gini': diversity.cand_pos_dist_gini,
    'cand_pos_dist_meandev': diversity.cand_pos_dist_meandev,
    'med_cands_summed': diversity.med_cands_summed,
    'vote_dist_mean': diversity.vote_dist_mean,
    'vote_dist_max': diversity.vote_dist_max,
    'vote_dist_med': diversity.vote_dist_med,
    'vote_dist_gini': diversity.vote_dist_gini,
    'vote_sqr_dist_mean': diversity.vote_sqr_dist_mean,
    'vote_sqr_dist_med': diversity.vote_sqr_dist_med,
    'vote_diversity_Karpov': diversity.vote_diversity_Karpov,
    'greedy_kKemenys_summed': diversity.greedy_kKemenys_summed,
    'greedy_2kKemenys_summed': diversity.greedy_2kKemenys_summed,
    'greedy_kKemenys_divk_summed': diversity.greedy_kKemenys_divk_summed,
    'polarization_1by2Kemenys': diversity.polarization_1by2Kemenys,
    'greedy_kmeans_summed': diversity.greedy_kmeans_summed,
    'support_pairs': diversity.support_pairs,
    'support_triplets': diversity.support_triplets,
    'support_votes': diversity.support_votes,
    'support_diversity_summed': diversity.support_diversity_summed,
    'support_diversity_normed_summed': diversity.support_diversity_normed_summed,
    'support_diversity_normed2_summed': diversity.support_diversity_normed2_summed,
    'support_diversity_normed3_summed': diversity.support_diversity_normed3_summed,
    'dist_to_Borda_mean': diversity.dist_to_Borda_mean,
    'dist_to_Kemeny_mean': diversity.dist_to_Kemeny_mean,
    'borda_spread': scores.borda_spread,
    'Agreement': diversity.agreement_index,
    'Diversity': diversity.diversity_index,
    'Polarization': diversity.polarization_index,

    'avg_distortion_from_guardians': distortion.avg_distortion_from_guardians,
    # unsupported feature
    'worst_distortion_from_guardians': distortion.worst_distortion_from_guardians,
    # unsupported feature
    'abstract': other.abstract,  # unsupported feature
    'distortion': distortion,  # unsupported feature
    'monotonicity_triplets': distortion.monotonicity_triplets,  # unsupported feature
    'partylist': partylist.partylistdistance,  # unsupported feature
    'pav_time': partylist.pav_time,  # unsupported feature
    'justified_ratio': other.justified_ratio,  # unsupported feature
    'rand_approx_pav_score': approx.get_rand_approx_pav_score,  # unsupported feature
    'min_dim': dimensionality.min_dim,  # unsupported feature
}


def get_global_feature(feature_id):
    """ Global feature depends on all instances """
    if feature_id in MAIN_GLOBAL_FEATUERS:
        return get_main_global_feature(feature_id)

    return {'clustering': clustering.clustering_v1,
            'clustering_kmeans': clustering.clustering_kmeans,
            'distortion_from_all': distortion.distortion_from_all,
            'id_vs_un': clustering.id_vs_un,
            'an_vs_st': clustering.an_vs_st,
            }.get(feature_id)


def get_local_feature(feature_id):
    """ Local feature depends only on a single instance """

    if feature_id in MAIN_LOCAL_FEATUERS:
        return get_main_local_feature(feature_id)
    elif feature_id in registered_approval_features:
        return registered_approval_features.get(feature_id)
    elif feature_id in registered_ordinal_features:
        return registered_ordinal_features.get(feature_id)
    else:
        raise ValueError(f'Incorrect feature id: {feature_id}')


def add_approval_feature(name, function):
    """
    Adds a new approval feature to the list of available approval features.

    :param name: name of the feature.
    :param function: function that calculates the feature.
    :return: None.
    """
    registered_approval_features[name] = function


def add_ordinal_feature(name, function):
    """
    Adds a new approval feature to the list of available approval features.

    :param name: name of the feature.
    :param function: function that calculates the feature.
    :return: None.
    """
    registered_ordinal_features[name] = function




from mapel.elections.cultures.mallows import get_mallows_vectors
from mapel.elections.cultures.matrices.single_peaked_matrices \
    import get_walsh_vectors, get_conitzer_vectors
from mapel.elections.cultures.params import *


def get_fake_multiplication(num_candidates, params, model):
    params['weight'] = 0.
    params['normphi'] = params['alpha']
    main_matrix = []
    if model == 'conitzer_path':
        main_matrix = get_conitzer_vectors(num_candidates).transpose()
    elif model == 'walsh_path':
        main_matrix = get_walsh_vectors(num_candidates).transpose()
    mallows_matrix = get_mallows_vectors(num_candidates, params).transpose()
    output = np.matmul(main_matrix, mallows_matrix).transpose()
    return output


def get_fake_vectors_single(fake_model_name, num_candidates):
    vectors = np.zeros([num_candidates, num_candidates])

    if fake_model_name == 'identity':
        for i in range(num_candidates):
            vectors[i][i] = 1

    elif fake_model_name == 'uniformity':
        for i in range(num_candidates):
            for j in range(num_candidates):
                vectors[i][j] = 1. / num_candidates

    elif fake_model_name == 'stratification':
        half = int(num_candidates / 2)
        for i in range(half):
            for j in range(half):
                vectors[i][j] = 1. / half
        for i in range(half, num_candidates):
            for j in range(half, num_candidates):
                vectors[i][j] = 1. / half

    elif fake_model_name == 'antagonism':
        for i in range(num_candidates):
            for _ in range(num_candidates):
                vectors[i][i] = 0.5
                vectors[i][num_candidates - i - 1] = 0.5

    return vectors


def get_fake_vectors_crate(num_candidates=None, fake_param=None):
    base_1 = get_fake_vectors_single('uniformity', num_candidates)
    base_2 = get_fake_vectors_single('identity', num_candidates)
    base_3 = get_fake_vectors_single('antagonism', num_candidates)
    base_4 = get_fake_vectors_single('stratification', num_candidates)

    return crate_combination(base_1, base_2, base_3, base_4, length=num_candidates,
                             alpha=fake_param)


def get_fake_convex(fake_model_name, num_candidates, num_voters, params, function_name):
    if fake_model_name == 'unid':
        base_1 = function_name('uniformity', num_candidates)
        base_2 = function_name('identity', num_candidates)
    elif fake_model_name == 'anid':
        base_1 = function_name('antagonism', num_candidates)
        base_2 = function_name('identity', num_candidates)
    elif fake_model_name == 'stid':
        base_1 = function_name('stratification', num_candidates)
        base_2 = function_name('identity', num_candidates)
    elif fake_model_name == 'anun':
        base_1 = function_name('antagonism', num_candidates)
        base_2 = function_name('uniformity', num_candidates)
    elif fake_model_name == 'stun':
        base_1 = function_name('stratification', num_candidates)
        base_2 = function_name('uniformity', num_candidates)
    elif fake_model_name == 'stan':
        base_1 = function_name('stratification', num_candidates)
        base_2 = function_name('antagonism', num_candidates)
    else:
        raise NameError('No such fake vectors/matrix!')

    return convex_combination(base_1, base_2, length=num_candidates, params=params)


def convex_combination(base_1, base_2, length=0, params=None):
    alpha = params['alpha']
    if base_1.ndim == 1:
        output = np.zeros([length])
        for i in range(length):
            output[i] = alpha * base_1[i] + (1 - alpha) * base_2[i]
    elif base_1.ndim == 2:
        output = np.zeros([length, length])
        for i in range(length):
            for j in range(length):
                output[i][j] = alpha * base_1[i][j] + (1 - alpha) * base_2[i][j]
    else:
        raise NameError('Unknown base!')
    return output


def crate_combination(base_1, base_2, base_3, base_4, length=0, alpha=None):
    alpha = alpha['alpha']
    vectors = np.zeros([length, length])
    for i in range(length):
        for j in range(length):
            vectors[i][j] = alpha[0] * base_1[i][j] + \
                            alpha[1] * base_2[i][j] + \
                            alpha[2] * base_3[i][j] + \
                            alpha[3] * base_4[i][j]

    return vectors


def get_fake_matrix_single(fake_model_name, num_candidates):
    matrix = np.zeros([num_candidates, num_candidates])

    if fake_model_name == 'identity':
        for i in range(num_candidates):
            for j in range(i + 1, num_candidates):
                matrix[i][j] = 1

    elif fake_model_name in {'uniformity', 'antagonism'}:
        for i in range(num_candidates):
            for j in range(num_candidates):
                if i != j:
                    matrix[i][j] = 0.5

    elif fake_model_name == 'stratification':
        for i in range(int(num_candidates / 2)):
            for j in range(int(num_candidates / 2), num_candidates):
                matrix[i][j] = 1
        for i in range(int(num_candidates / 2)):
            for j in range(int(num_candidates / 2)):
                if i != j:
                    matrix[i][j] = 0.5
        for i in range(int(num_candidates / 2), num_candidates):
            for j in range(int(num_candidates / 2), num_candidates):
                if i != j:
                    matrix[i][j] = 0.5

    return matrix


def get_fake_borda_vector(fake_model_name, num_candidates, num_voters):
    borda_vector = np.zeros([num_candidates])

    m = num_candidates
    n = num_voters

    if fake_model_name == 'identity':
        for i in range(m):
            borda_vector[i] = n * (m - 1 - i)

    elif fake_model_name in {'uniformity', 'antagonism'}:
        for i in range(m):
            borda_vector[i] = n * (m - 1) / 2

    elif fake_model_name == 'stratification':
        for i in range(int(m / 2)):
            borda_vector[i] = n * (m - 1) * 3 / 4
        for i in range(int(m / 2), m):
            borda_vector[i] = n * (m - 1) / 4

    return borda_vector



#!/usr/bin/env python
import csv
import logging
import warnings
from abc import ABCMeta, abstractmethod
from multiprocessing import Process
from time import sleep
import ast
import time
from tqdm import tqdm

from mapel.elections.objects.ElectionFamily import ElectionFamily
from mapel.elections.objects.OrdinalElection import OrdinalElection
from mapel.elections.objects.ApprovalElection import ApprovalElection
import mapel.elections.distances_ as metr
import mapel.elections.other.rules as rules
import mapel.elections.features_ as features
from mapel.core.objects.Experiment import Experiment
import mapel.core.printing as pr
from mapel.core.utils import *
from mapel.core.glossary import *

import mapel.core.persistence.experiment_exports as exports

try:
    from sklearn.manifold import MDS
    from sklearn.manifold import TSNE
    from sklearn.manifold import SpectralEmbedding
    from sklearn.manifold import LocallyLinearEmbedding
    from sklearn.manifold import Isomap
except ImportError as error:
    MDS = None
    TSNE = None
    SpectralEmbedding = None
    LocallyLinearEmbedding = None
    Isomap = None
    print(error)


class ElectionExperiment(Experiment):
    __metaclass__ = ABCMeta
    """Abstract set of instances."""

    @abstractmethod
    def add_folders_to_experiment(self):
        pass

    @abstractmethod
    def prepare_instances(self):
        pass

    @abstractmethod
    def add_instance(self):
        pass

    @abstractmethod
    def add_feature(self, name, function):
        pass

    @abstractmethod
    def add_culture(self, name, function):
        pass

    def __init__(self, is_shifted=False, **kwargs):
        self.is_shifted = is_shifted
        self.default_num_candidates = 10
        self.default_num_voters = 100
        self.default_committee_size = 1
        self.all_winning_committees = {}
        super().__init__(**kwargs)

    def __getattr__(self, attr):
        if attr == 'elections':
            return self.instances
        elif attr == 'num_elections':
            return self.num_instances
        else:
            return self.__dict__[attr]

    def __setattr__(self, name, value):
        if name == "elections":
            self.instances = value
        elif name == "num_elections":
            self.num_instances = value
        else:
            self.__dict__[name] = value

    def prepare_matrices(self):
        path = os.path.join(os.getcwd(), "experiments", self.experiment_id, "matrices")
        for file_name in os.listdir(path):
            os.remove(os.path.join(path, file_name))

        for election_id in self.elections:
            matrix = self.elections[election_id].votes_to_positionwise_matrix()
            file_name = election_id + ".csv"
            path = os.path.join(os.getcwd(), "experiments", self.experiment_id,
                                "matrices", file_name)

            with open(path, 'w', newline='') as csv_file:

                writer = csv.writer(csv_file, delimiter=';')
                header = [str(i) for i in range(self.elections[election_id].num_candidates)]
                writer.writerow(header)
                for row in matrix:
                    writer.writerow(row)

    def add_instances_to_experiment(self):
        instances = {}

        for family_id in self.families:
            single = self.families[family_id].single
            ids = []
            for j in range(self.families[family_id].size):
                instance_id = get_instance_id(single, family_id, j)
                if self.instance_type == 'ordinal':
                    instance = OrdinalElection(self.experiment_id, instance_id,
                                               is_imported=True,
                                               fast_import=self.fast_import,
                                               with_matrix=self.with_matrix,
                                               label=self.families[family_id].label)
                elif self.instance_type == 'approval':
                    instance = ApprovalElection(self.experiment_id, instance_id,
                                                is_imported=True,
                                                fast_import=self.fast_import,
                                                label=self.families[family_id].label)
                else:
                    instance = None

                instances[instance_id] = instance
                ids.append(str(instance_id))

            self.families[family_id].election_ids = ids

        return instances

    def set_default_num_candidates(self, num_candidates: int) -> None:
        """ Set default number of candidates """
        self.default_num_candidates = num_candidates

    def set_default_num_voters(self, num_voters: int) -> None:
        """ Set default number of voters """
        self.default_num_voters = num_voters

    def set_default_committee_size(self, committee_size: int) -> None:
        """ Set default size of the committee """
        self.default_committee_size = committee_size

    def add_election(self,
                     culture_id="none",
                     params=None,
                     label=None,
                     color="black",
                     alpha=1.,
                     show=True,
                     marker='x',
                     ms=20,
                     starting_from=0,
                     size=1,
                     num_candidates=None,
                     num_voters=None,
                     election_id=None,  # deprecated
                     instance_id=None):
        """ Add election to the experiment """

        if instance_id is None:
            instance_id = election_id

        if num_candidates is None:
            num_candidates = self.default_num_candidates

        if num_voters is None:
            num_voters = self.default_num_voters

        return self.add_family(culture_id=culture_id,
                               params=params,
                               size=size,
                               label=label,
                               color=color,
                               alpha=alpha,
                               show=show,
                               marker=marker,
                               ms=ms,
                               starting_from=starting_from,
                               family_id=instance_id,
                               num_candidates=num_candidates,
                               num_voters=num_voters,
                               single=True)

    def add_family(self,
                   culture_id: str = "none",
                   params: dict = None,
                   size: int = 1,
                   label: str = None,
                   color: str = "black",
                   alpha: float = 1.,
                   show: bool = True,
                   marker: str = 'o',
                   ms: int = 20,
                   starting_from: int = 0,
                   num_candidates: int = None,
                   num_voters: int = None,
                   family_id: str = None,
                   single: bool = False,
                   path: dict = None,
                   election_id: str = None,
                   **kwargs) -> list:
        """ Add family of elections to the experiment """

        if election_id is not None:
            family_id = election_id

        if num_candidates is None:
            num_candidates = self.default_num_candidates

        if num_voters is None:
            num_voters = self.default_num_voters

        if self.families is None:
            self.families = {}

        if family_id is None:
            family_id = culture_id + '_' + str(num_candidates) + '_' + str(num_voters)
            if culture_id in {'urn'} and params['alpha'] is not None:
                family_id += '_' + str(float(params['alpha']))
            elif culture_id in {'mallows'} and params['phi'] is not None:
                family_id += '_' + str(float(params['phi']))
            elif culture_id in {'norm-mallows', 'norm-mallows_matrix'} \
                    and params['normphi'] is not None:
                family_id += '_' + str(float(params['normphi']))

        elif label is None:
            label = family_id
        self.families[family_id] = ElectionFamily(culture_id=culture_id,
                                                  family_id=family_id,
                                                  params=params,
                                                  label=label,
                                                  color=color,
                                                  alpha=alpha,
                                                  show=show,
                                                  size=size,
                                                  marker=marker,
                                                  ms=ms,
                                                  starting_from=starting_from,
                                                  num_candidates=num_candidates,
                                                  num_voters=num_voters,
                                                  path=path,
                                                  single=single,
                                                  instance_type=self.instance_type,
                                                  **kwargs)

        self.num_families = len(self.families)
        self.num_elections = sum([self.families[family_id].size for family_id in self.families])
        self.main_order = [i for i in range(self.num_elections)]

        new_instances = self.families[family_id].prepare_family(
            is_exported=self.is_exported,
            experiment_id=self.experiment_id,
            instance_type=self.instance_type)

        for instance_id in new_instances:
            self.instances[instance_id] = new_instances[instance_id]

        self.families[family_id].instance_ids = list(new_instances.keys())

        if self.is_exported:
            self.update_map_csv()

        return list(new_instances.keys())

    def add_existing_family_from_dir(self,
                                     dir=None,
                                     culture_id=None,
                                     **kwargs) -> list:
        """ Add family of elections to the experiment """
        if dir is None:
            logging.warning('dir not specified')
        if culture_id is None:
            logging.warning('culture_id not specified')

        # Copy instances from dir to /elections

        directory_in = dir
        directory_out = os.path.join(os.getcwd(),
                                     "experiments",
                                     self.experiment_id,
                                     'elections')

        # List all files in the given directory
        files = [f for f in os.listdir(directory_in) if
                 os.path.isfile(os.path.join(directory_in, f))]

        # Sort the files for consistency
        files.sort()

        # Rename each file
        for idx, file_name in enumerate(files):
            # Create the new file name
            # new_file_name = f"{culture_id}_{idx}{os.path.splitext(file_name)[1]}"
            new_file_name = f"{culture_id}_{idx}.soc"

            # Form the full old and new paths
            old_path = os.path.join(directory_in, file_name)
            new_path = os.path.join(directory_out, new_file_name)

            # Rename the file
            os.rename(old_path, new_path)
            print(f"Renaming: {file_name} -> {new_file_name}")

        return self.add_family(culture_id=culture_id, **kwargs)

    def update_map_csv(self):

        """ Import controllers from a file """

        families = {}

        path = os.path.join(os.getcwd(), 'experiments', self.experiment_id, 'map.csv')
        with open(path, 'w', newline='') as csv_file:
            writer = csv.writer(csv_file, delimiter=';')

            all_fields = ['size',
                          'num_candidates',
                          'num_voters',
                          'culture_id',
                          'params',
                          'family_id',
                          'label',
                          'color',
                          'alpha',
                          'marker',
                          'ms',
                          'path'
                          ]

            writer.writerow(all_fields)

            for family in self.families.values():
                all_values = [family.size,
                              family.num_candidates,
                              family.num_voters,
                              family.culture_id,
                              family.params,
                              family.family_id,
                              family.label,
                              family.color,
                              family.alpha,
                              family.marker,
                              family.ms,
                              family.path]

                writer.writerow(all_values)

        return families

    def add_empty_family(self,
                         culture_id: str = "none",
                         label: str = None,
                         color: str = "black",
                         alpha: float = 1.,
                         show: bool = True,
                         marker: str = 'o',
                         num_candidates: int = None,
                         num_voters: int = None,
                         family_id: str = None):

        if label is None:
            label = family_id
        self.families[family_id] = ElectionFamily(culture_id=culture_id,
                                                  family_id=family_id,
                                                  label=label,
                                                  color=color,
                                                  alpha=alpha,
                                                  show=show,
                                                  size=0,
                                                  marker=marker,
                                                  num_candidates=num_candidates,
                                                  num_voters=num_voters,
                                                  instance_type=self.instance_type)

        self.families[family_id].prepare_family(
            is_exported=self.is_exported,
            experiment_id=self.experiment_id)

        return self.families[family_id]

    def prepare_elections(self, store_points=False, is_aggregated=True):
        """ Prepare elections for a given experiment """

        self.store_points = store_points
        self.is_aggregated = is_aggregated

        if self.instances is None:
            self.instances = {}

        for family_id in tqdm(self.families, desc="Preparing instances"):

            new_instances = self.families[family_id].prepare_family(
                is_exported=self.is_exported,
                experiment_id=self.experiment_id,
                store_points=store_points,
                is_aggregated=is_aggregated,
                instance_type=self.instance_type,
            )

            for instance_id in new_instances:
                self.instances[instance_id] = new_instances[instance_id]

    def compute_winners(self, method=None, num_winners=1):
        for election_id in self.elections:
            self.elections[election_id].compute_winners(method=method, num_winners=num_winners)

    def compute_alternative_winners(self, method=None, num_winners=None, num_parties=None):
        for election_id in self.elections:
            for party_id in range(num_parties):
                self.elections[election_id].compute_alternative_winners(
                    method=method, party_id=party_id, num_winners=num_winners)

    def compute_distances(self,
                          distance_id: str = None,
                          num_processes: int = 1,
                          self_distances: bool = False,
                          ids = None,
                          **kwargs) -> None:
        """ Compute distances between elections (using processes) """

        if distance_id is None:
            distance_id = self.distance_id

        self.distance_id = distance_id

        if '-approvalwise' in distance_id:
            for election in self.elections.values():
                election.votes_to_approvalwise_vector()
        elif '-coapproval_frequency' in distance_id:
            for election in self.elections.values():
                election.votes_to_coapproval_frequency_vectors(**kwargs)
        elif '-voterlikeness' in distance_id:
            for election in self.elections.values():
                election.votes_to_voterlikeness_matrix(**kwargs)
        elif '-candidatelikeness' in distance_id:
            for election in self.elections.values():
                election.votes_to_candidatelikeness_sorted_vectors()
        elif '-pairwise' in distance_id:
            for election in self.elections.values():
                election.votes_to_pairwise_matrix()

        matchings = {election_id: {} for election_id in self.elections}
        distances = {election_id: {} for election_id in self.elections}
        times = {election_id: {} for election_id in self.elections}

        if ids is None:
            ids = []
            for i, election_1 in enumerate(self.elections):
                for j, election_2 in enumerate(self.elections):
                    if i < j or (i == j and self_distances):
                        ids.append((election_1, election_2))

        num_distances = len(ids)

        if self.experiment_id == 'virtual' or num_processes == 1:
            metr.run_single_process(self, ids, distances, times, matchings)

        else:
            processes = []
            for process_id in range(num_processes):
                print(f'Starting process: {process_id}')
                sleep(0.1)
                start = int(process_id * num_distances / num_processes)
                stop = int((process_id + 1) * num_distances / num_processes)
                instances_ids = ids[start:stop]

                process = Process(target=metr.run_multiple_processes, args=(self,
                                                                            instances_ids,
                                                                            distances,
                                                                            times,
                                                                            matchings,
                                                                            process_id))
                process.start()
                processes.append(process)

            for process in processes:
                process.join()

            distances = {instance_id: {} for instance_id in self.instances}
            times = {instance_id: {} for instance_id in self.instances}

            for process_id in range(num_processes):

                file_name = f'{distance_id}_p{process_id}.csv'
                path = os.path.join(os.getcwd(),
                                    "experiments",
                                    self.experiment_id,
                                    "distances",
                                    file_name)

                with open(path, 'r', newline='') as csv_file:
                    reader = csv.DictReader(csv_file, delimiter=';')

                    for row in reader:
                        distances[row['instance_id_1']][row['instance_id_2']] = \
                            float(row['distance'])
                        times[row['instance_id_1']][row['instance_id_2']] = \
                            float(row['time'])

        if self.is_exported:
            exports.export_distances_to_file(self,
                                             distance_id,
                                             distances,
                                             times,
                                             self_distances,
                                             ids=ids)

        self.distances = distances
        self.times = times
        self.matchings = matchings

    def get_election_id_from_model_name(self, culture_id: str) -> str:
        for family_id in self.families:
            if self.families[family_id].culture_id == culture_id:
                return family_id

    def print_matrix(self, **kwargs):
        pr.print_matrix(experiment=self, **kwargs)

    def import_controllers(self):
        """ Import controllers from a file """

        families = {}

        path = os.path.join(os.getcwd(), 'experiments', self.experiment_id, 'map.csv')
        with open(path, 'r') as file_:

            header = [h.strip() for h in file_.readline().split(';')]
            reader = csv.DictReader(file_, fieldnames=header, delimiter=';')

            all_num_candidates = []
            all_num_voters = []

            starting_from = 0
            for row in reader:

                culture_id = None
                params = None
                size = None
                num_candidates = None
                num_voters = None
                family_id = None

                print_params = {}

                if 'culture_id' in row.keys():
                    culture_id = str(row['culture_id']).strip()

                if 'family_id' in row.keys():
                    family_id = str(row['family_id'])

                if 'params' in row.keys():
                    params = ast.literal_eval(str(row['params']))

                if 'size' in row.keys():
                    size = int(row['size'])

                if 'num_candidates' in row.keys():
                    num_candidates = int(row['num_candidates'])

                if 'num_voters' in row.keys():
                    num_voters = int(row['num_voters'])

                if 'path' in row.keys():
                    path = ast.literal_eval(str(row['path']))

                if 'label' in row.keys():
                    print_params['label'] = str(row['label'])
                if 'alpha' in row.keys():
                    print_params['alpha'] = float(row['alpha'])
                if 'marker' in row.keys():
                    print_params['marker'] = str(row['marker']).strip()
                if 'ms' in row.keys():
                    print_params['ms'] = int(row['ms'])
                if 'color' in row.keys():
                    print_params['color'] = str(row['color']).strip()

                single = size == 1

                families[family_id] = ElectionFamily(culture_id=culture_id,
                                                     family_id=family_id,
                                                     params=params,
                                                     size=size,
                                                     starting_from=starting_from,
                                                     num_candidates=num_candidates,
                                                     num_voters=num_voters,
                                                     path=path,
                                                     single=single,
                                                     **print_params
                                                     )
                starting_from += size

                all_num_candidates.append(num_candidates)
                all_num_voters.append(num_voters)

            check_if_all_equal(all_num_candidates, 'num_candidates')
            check_if_all_equal(all_num_voters, 'num_voters')

            self.num_families = len(families)
            self.num_elections = sum([families[family_id].size for family_id in families])
            self.main_order = [i for i in range(self.num_elections)]

        return families

    def compute_feature(self,
                        feature_id: str = None,
                        feature_params=None,
                        overwrite=False,
                        saveas=None,
                        **kwargs) -> dict:

        if feature_params is None:
            feature_params = {}

        if feature_id in ['priceability', 'core', 'ejr']:
            feature_long_id = f'{feature_id}_{feature_params["rule"]}'
        elif feature_id in ['distortion', 'monotonicity']:
            feature_long_id = f'{feature_id}_{self.embedding_id}'
        else:
            feature_long_id = feature_id

        num_iterations = 1
        if 'num_iterations' in feature_params:
            num_iterations = feature_params['num_iterations']

        if feature_id == 'ejr':
            feature_dict = {'value': {}, 'time': {}, 'ejr': {}, 'pjr': {}, 'jr': {}, 'pareto': {}}
        else:
            feature_dict = {'value': {}, 'time': {}}

        if feature_id in MAIN_GLOBAL_FEATUERS or feature_id in ELECTION_GLOBAL_FEATURES:

            feature = features.get_global_feature(feature_id)

            values = feature(self, election_ids=list(self.instances), **kwargs)

            for instance_id in tqdm(self.instances, desc='Computing feature'):
                feature_dict['value'][instance_id] = values[instance_id]
                if values[instance_id] is None:
                    feature_dict['time'][instance_id] = None
                else:
                    feature_dict['time'][instance_id] = 0

        else:
            feature = features.get_local_feature(feature_id)

            for instance_id in tqdm(self.instances):
                instance = self.elections[instance_id]

                start = time.time()
                solution = None
                for _ in range(num_iterations):

                    if feature_id in ['monotonicity_1',
                                      'monotonicity_triplets']:
                        value = feature(self, instance)

                    elif feature_id in {'avg_distortion_from_guardians',
                                        'worst_distortion_from_guardians',
                                        'distortion_from_all',
                                        'distortion_from_top_100'}:
                        value = feature(self, instance_id)
                    elif feature_id in ['ejr',
                                        'core',
                                        'pareto',
                                        'priceability',
                                        'cohesiveness']:
                        value = instance.get_feature(feature_id, feature_long_id,
                                                     feature_params=feature_params)
                    else:
                        solution = instance.get_feature(feature_id, feature_long_id,
                                                        overwrite=overwrite, **kwargs)
                        value = None

                total_time = time.time() - start
                total_time /= num_iterations

                if solution is not None:
                    if type(solution) is dict:
                        for key in solution:
                            if key not in feature_dict:
                                feature_dict[key] = {}
                            feature_dict[key][instance_id] = solution[key]
                    else:
                        feature_dict['value'][instance_id] = solution
                    feature_dict['time'][instance_id] = total_time
                else:
                    feature_dict['value'][instance_id] = value
                    feature_dict['time'][instance_id] = total_time

        if saveas is None:
            saveas = feature_long_id

        if self.is_exported:
            exports.export_feature_to_file(self,
                                           feature_id=feature_id,
                                           feature_dict=feature_dict,
                                           saveas=saveas)

        self.features[saveas] = feature_dict
        return feature_dict

    def compute_rules(self, list_of_rules,
                      committee_size: int = 10,
                      resolute: bool = False) -> None:
        for rule_name in list_of_rules:
            print('Computing', rule_name)
            if rule_name in NOT_ABCVOTING_RULES:
                rules.compute_not_abcvoting_rule(experiment=self,
                                                 rule_name=rule_name,
                                                 committee_size=committee_size,
                                                 resolute=resolute)
            else:
                rules.compute_abcvoting_rule(experiment=self, rule_name=rule_name,
                                             committee_size=committee_size,
                                             resolute=resolute)

    def import_committees(self, list_of_rules):
        for rule_name in list_of_rules:
            self.all_winning_committees[rule_name] = rules.import_committees_from_file(
                experiment_id=self.experiment_id, rule_name=rule_name)

    def add_election_to_family(self, election=None, family_id=None):
        self.instances[election.instance_id] = election
        self.families[family_id].add_election(election)


def check_if_all_equal(values, subject):
    if any(x != values[0] for x in values):
        text = f'Not all {subject} values are equal!'
        warnings.warn(text)

# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 22.10.2021 #
# # # # # # # # # # # # # # # #



import numpy as np


def num_of_diff_votes(election):
    if election.fake:
        return {'value': None}

    str_votes = [str(vote) for vote in election.votes]
    return {'value': len(set(str_votes))}


def voterlikeness_sqrt(election):
    if election.fake:
        return {'value': None}

    vectors = election.votes_to_voterlikeness_vectors()
    score = 0.
    for vector in vectors:
        for value in vector:
            score += value**0.5
    return {'value': score}


def voterlikeness_harmonic(election):
    if election.fake:
        return {'value': None}

    vectors = election.votes_to_voterlikeness_vectors()
    score = 0.
    for vector in vectors:
        vector = sorted(vector)
        for pos, value in enumerate(vector):
            score += 1/(pos+2)*value
    return {'value': score}


def borda_diversity(election):
    if election.fake:
        return {'value': None}

    vector = np.array(election.votes_to_bordawise_vector())
    return {'value': np.std(vector)}



import logging
import os
from typing import Callable, List
from itertools import combinations, permutations

from mapel.core.matchings import *
from mapel.elections.objects.OrdinalElection import OrdinalElection
import mapel.core.utils as utils
from mapel.core.inner_distances import swap_distance
import mapel.elections.distances.ilp_isomorphic as ilp_iso

import mapel.elections.distances.lp as lp

try:
    import mapel.elections.distances.cppdistances as cppd
except:
    logging.warning("The quick C++ procedures for computing the swap and "
                    "Spearman distance is unavailable: using the (slow) python one instead")


# MAIN DISTANCES
def compute_pos_swap_distance(election_1: OrdinalElection, election_2: OrdinalElection,
                              inner_distance: Callable) -> (float, list):
    """ Compute Positionwise distance between ordinal elections """
    cost_table = get_matching_cost_positionwise(election_1, election_2, inner_distance)
    obj_val, matching = solve_matching_vectors(cost_table)
    cost_table = get_matching_cost_pos_swap(election_1, election_2, matching)
    return solve_matching_vectors(cost_table)


def compute_positionwise_distance(election_1: OrdinalElection, election_2: OrdinalElection,
                                  inner_distance: Callable) -> (float, list):
    """ Compute Positionwise distance between ordinal elections """
    cost_table = get_matching_cost_positionwise(election_1, election_2, inner_distance)
    return solve_matching_vectors(cost_table)


def compute_agg_voterlikeness_distance(election_1: OrdinalElection, election_2: OrdinalElection,
                                       inner_distance: Callable) -> (float, list):
    """ Compute Aggregated-Voterlikeness distance between ordinal elections """
    vector_1, num_possible_scores = election_1.votes_to_agg_voterlikeness_vector()
    vector_2, _ = election_2.votes_to_agg_voterlikeness_vector()
    return inner_distance(vector_1, vector_2, num_possible_scores)


def compute_bordawise_distance(election_1: OrdinalElection, election_2: OrdinalElection,
                               inner_distance: Callable) -> (float, list):
    """ Compute Bordawise distance between ordinal elections """
    vector_1 = election_1.votes_to_bordawise_vector()
    vector_2 = election_2.votes_to_bordawise_vector()
    return inner_distance(vector_1, vector_2), None


def compute_pairwise_distance(election_1: OrdinalElection, election_2: OrdinalElection,
                              inner_distance: Callable) -> (float, list):
    """ Compute Pairwise distance between ordinal elections """
    length = election_1.num_candidates
    matrix_1 = election_1.votes_to_pairwise_matrix()
    matrix_2 = election_2.votes_to_pairwise_matrix()
    return solve_matching_matrices(matrix_1, matrix_2, length, inner_distance), None


def compute_voterlikeness_distance(election_1: OrdinalElection, election_2: OrdinalElection,
                                   inner_distance: Callable) -> (float, list):
    """ Compute Voterlikeness distance between elections """
    length = election_1.num_voters
    matrix_1 = election_1.votes_to_voterlikeness_matrix()
    matrix_2 = election_2.votes_to_voterlikeness_matrix()
    return solve_matching_matrices(matrix_1, matrix_2, length, inner_distance), None


# DEPRECATED
def compute_swap_distance_bf(election_1: OrdinalElection,
                             election_2: OrdinalElection) -> (int, list):
    """ Compute swap distance between elections (using brute force) """
    obj_values = []
    for mapping in permutations(range(election_1.num_candidates)):
        cost_table = get_matching_cost_swap_bf(election_1, election_2, mapping)
        obj_values.append(solve_matching_vectors(cost_table)[0])
    return min(obj_values), None


def compute_swap_distance(election_1: OrdinalElection,
                          election_2: OrdinalElection) -> (int, list):
    """ Compute swap distance between elections (using the C++ extension) """
    if not utils.is_module_loaded("mapel.elections.distances.cppdistances"):
        return compute_swap_distance_ilp_py(election_1, election_2), None

    if election_1.num_candidates < election_2.num_candidates:
        swapd = cppd.tswapd(election_1.votes.tolist(),
                           election_2.votes.tolist())
    elif election_1.num_candidates > election_2.num_candidates:
        swapd = cppd.tswapd(election_2.votes.tolist(),
                           election_1.votes.tolist())
    else:
        swapd = cppd.swapd(election_1.votes.tolist(),
                           election_2.votes.tolist())

    return swapd, None


def compute_truncated_swap_distance(election_1: OrdinalElection,
                          election_2: OrdinalElection) -> (int, list):
    """ Compute truncated swap distance between elections """
    obj_values = []
    for mapping in permutations(range(election_1.num_candidates)):
        cost_table = get_matching_cost_truncated_swap_bf(election_1, election_2, mapping)
        obj_values.append(solve_matching_vectors(cost_table)[0])
    return min(obj_values), None


def compute_spearman_distance(election_1: OrdinalElection,
                              election_2: OrdinalElection) -> (int, list):
    """ Compute Spearman distance between elections (using the C++ extension) """
    if not utils.is_module_loaded("mapel.elections.distances.cppdistances"):
        return compute_spearman_distance_ilp_py(election_1, election_2), None
    speard = cppd.speard(election_1.votes.tolist(),
                         election_2.votes.tolist())
    return speard, None


def compute_spearman_distance_ilp_py(election_1: OrdinalElection,
                                     election_2: OrdinalElection) -> (int, list):
    """ Compute Spearman distance between elections """
    votes_1 = election_1.votes
    votes_2 = election_2.votes
    params = {'voters': election_1.num_voters,
              'candidates': election_1.num_candidates}

    file_name = f'{np.random.random()}.lp'
    path = os.path.join(os.getcwd(), "trash", file_name)
    ilp_iso.generate_ilp_spearman_distance(path, votes_1, votes_2, params)
    objective_value = ilp_iso.solve_ilp_distance(path)
    ilp_iso.remove_lp_file(path)
    objective_value = int(round(objective_value, 0))
    return objective_value, None


def compute_swap_distance_ilp_py(election_1: OrdinalElection,
                                 election_2: OrdinalElection) -> (int, list):
    """ Compute Spearman distance between elections """
    votes_1 = election_1.votes
    votes_2 = election_2.votes
    params = {'voters': election_1.num_voters,
              'candidates': election_1.num_candidates}

    file_name = f'{np.random.random()}.lp'
    path = os.path.join(os.getcwd(), 'trash', file_name)
    ilp_iso.generate_ilp_swap_distance(path, votes_1, votes_2, params)
    objective_value = ilp_iso.solve_ilp_distance_swap(path, votes_1, votes_2, params)
    ilp_iso.remove_lp_file(path)
    objective_value = int(round(objective_value, 0))
    return objective_value, None


def compute_discrete_distance(election_1: OrdinalElection,
                              election_2: OrdinalElection) -> (int, list):
    """ Compute Discrete distance between elections """
    return election_1.num_voters - compute_voter_subelection(election_1, election_2), None


# SUBELECTIONS #
def compute_voter_subelection(election_1: OrdinalElection, election_2: OrdinalElection) -> int:
    """ Compute Voter-Subelection """
    return lp.solve_lp_voter_subelection(election_1, election_2)


def compute_candidate_subelection(election_1: OrdinalElection, election_2: OrdinalElection) -> int:
    """ Compute Candidate-Subelection """
    file_name = f'{np.random.random()}.lp'
    path = os.path.join(os.getcwd(), 'trash', file_name)
    objective_value = lp.solve_lp_candidate_subelections(path, election_1, election_2)
    lp.remove_lp_file(path)
    return objective_value


# HELPER FUNCTIONS #
def get_matching_cost_pos_swap(election_1: OrdinalElection, election_2: OrdinalElection,
                               matching) -> List[list]:
    """ Return: Cost table """
    votes_1 = election_1.votes
    votes_2 = election_2.votes
    size = election_1.num_voters
    return [[swap_distance(votes_1[i], votes_2[j], matching=matching) for i in range(size)]
            for j in range(size)]


def get_matching_cost_positionwise(election_1: OrdinalElection, election_2: OrdinalElection,
                                   inner_distance: Callable) -> List[list]:
    """ Return: Cost table """

    vectors_1 = election_1.get_vectors()
    vectors_2 = election_2.get_vectors()
    size = election_1.num_candidates
    return [[inner_distance(vectors_1[i], vectors_2[j]) for i in range(size)] for j in range(size)]


def get_matching_cost_swap_bf(election_1: OrdinalElection, election_2: OrdinalElection,
                              mapping):
    """ Return: Cost table """
    cost_table = np.zeros([election_1.num_voters, election_1.num_voters])

    election_1.get_potes()
    election_2.get_potes()

    for v1 in range(election_1.num_voters):
        for v2 in range(election_2.num_voters):
            swap_distance = 0
            for i, j in combinations(election_1.potes[0], 2):
                if (election_1.potes[v1][i] > election_1.potes[v1][j] and
                    election_2.potes[v2][mapping[i]] < election_2.potes[v2][mapping[j]]) or \
                        (election_1.potes[v1][i] < election_1.potes[v1][j] and
                         election_2.potes[v2][mapping[i]] > election_2.potes[v2][mapping[j]]):
                    swap_distance += 1
            cost_table[v1][v2] = swap_distance
    return cost_table


def get_matching_cost_truncated_swap_bf(election_1: OrdinalElection,
                                        election_2: OrdinalElection,
                                        mapping):
    """ Return: Cost table """
    cost_table = np.zeros([election_1.num_voters, election_1.num_voters])

    election_1.get_potes()
    election_2.get_potes()

    for v1 in range(election_1.num_voters):
        for v2 in range(election_2.num_voters):
            swap_distance = 0
            for i, j in combinations(election_1.potes[0], 2):
                if (election_1.potes[v1][i] > election_1.potes[v1][j] and
                    election_2.potes[v2][mapping[i]] < election_2.potes[v2][mapping[j]]) or \
                        (election_1.potes[v1][i] < election_1.potes[v1][j] and
                         election_2.potes[v2][mapping[i]] > election_2.potes[v2][mapping[j]]):
                    swap_distance += 1
            cost_table[v1][v2] = swap_distance
    return cost_table


def compute_blank_distance(election_1: OrdinalElection,
                           election_2: OrdinalElection) -> (int, list):
    return 1, None



#!/usr/bin/env python
from abc import ABC

from mapel.core.matchings import solve_matching_vectors
from mapel.elections.objects.ElectionExperiment import ElectionExperiment
from mapel.elections.other import pabulib
from mapel.core.utils import *
import numpy as np
import csv

import mapel.elections.cultures_ as cultures
import mapel.elections.features_ as features
import mapel.elections.distances_ as distances
from tqdm import tqdm

try:
    from sklearn.manifold import MDS
    from sklearn.manifold import TSNE
    from sklearn.manifold import SpectralEmbedding
    from sklearn.manifold import LocallyLinearEmbedding
    from sklearn.manifold import Isomap
except ImportError as error:
    MDS = None
    TSNE = None
    SpectralEmbedding = None
    LocallyLinearEmbedding = None
    Isomap = None
    print(error)


class ApprovalElectionExperiment(ElectionExperiment, ABC):
    """ Abstract set of approval elections."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def add_culture(self, name, function):
        cultures.add_approval_culture(name, function)

    def add_feature(self, name, function):
        features.add_approval_feature(name, function)

    def add_distance(self, name, function):
        distances.add_approval_distance(name, function)

    def compute_distance_between_rules(self, list_of_rules=None, printing=False,
                                       distance_id=None, committee_size=10):

        self.import_committees(list_of_rules=list_of_rules)

        path = os.path.join(os.getcwd(), "experiments", f'{self.experiment_id}', '..',
                            'rules_output', 'distances', f'{distance_id}.csv')

        with open(path, 'w', newline='') as csv_file:
            writer = csv.writer(csv_file, delimiter=';')
            writer.writerow(["election_id_1", "election_id_2", "distance", "time"])

            for i, r1 in enumerate(list_of_rules):
                for j, r2 in enumerate(list_of_rules):
                    if i < j:
                        if printing:
                            print(r1, r2)
                        all_distance = []
                        for election_id in self.elections:
                            com1 = self.all_winning_committees[r1][
                                election_id][0]
                            com2 = self.all_winning_committees[r2][
                                election_id][0]

                            if distance_id == 'discrete':
                                distance = len(com1.symmetric_difference(com2))
                            elif distance_id == 'wrong':
                                distance = 1 - len(com1.intersection(com2)) / len(com1.union(com2))
                            elif distance_id in ['hamming', 'jaccard']:
                                cand_dist = np.zeros([committee_size, committee_size])

                                self.elections[election_id].compute_reverse_approvals()
                                for k1, c1 in enumerate(com1):
                                    for k2, c2 in enumerate(com2):

                                        ac1 = self.elections[election_id].reverse_approvals[c1]
                                        ac2 = self.elections[election_id].reverse_approvals[c2]
                                        if distance_id == 'hamming':
                                            cand_dist[k1][k2] = len(ac1.symmetric_difference(ac2))
                                        elif distance_id == 'jaccard':
                                            if len(ac1.union(ac2)) != 0:
                                                cand_dist[k1][k2] = 1 - len(
                                                    ac1.intersection(ac2)) / len(ac1.union(ac2))
                                distance, _ = solve_matching_vectors(cand_dist)
                                distance /= committee_size
                                if printing:
                                    print(distance)
                            all_distance.append(distance)
                        mean = sum(all_distance) / self.num_elections
                        writer.writerow([r1, r2, mean, 0.])

    def compute_rule_features(self,
                              feature_id=None,
                              list_of_rules=None,
                              feature_params=None,
                              **kwargs):
        if feature_params is None:
            feature_params = {}

        self.import_committees(list_of_rules=list_of_rules)

        for election_id in self.elections:
            self.elections[election_id].winning_committee = {}

        for r in list_of_rules:
            for election_id in self.elections:
                self.elections[election_id].winning_committee[r] = \
                    self.all_winning_committees[r][election_id][0]

        for rule in tqdm(list_of_rules):
            feature_params['rule'] = rule
            self.compute_feature(feature_id=feature_id, feature_params=feature_params, **kwargs)

    def print_latex_table(self,
                          feature_id=None,
                          column_id='value',
                          list_of_rules=None,
                          list_of_models=None):

        features = {}
        for rule in list_of_rules:
            features[rule] = self.import_feature(feature_id, column_id=column_id, rule=rule)

        results = {}
        for model in list_of_models:
            results[model] = {}
            for rule in list_of_rules:
                feature = features[rule]
                total_value = 0
                ctr = 0
                for instance in feature:
                    if model in instance:
                        total_value += feature[instance]
                        ctr += 1
                avg_value = round(total_value / ctr, 2)
                results[model][rule] = avg_value

        print("\\toprule")
        print("rule", end=" ")
        for model in list_of_models:
            print(f'& {model}', end=" ")
        print("\\\\ \\midrule")

        for rule in list_of_rules:
            print(rule, end=" ")
            for model in list_of_models:
                print(f'& {results[model][rule]}', end=" ")
            # print("")
            print("\\\\ \\midrule")

    def print_latex_multitable(self,
                               features_id=None,
                               columns_id=None,
                               list_of_rules=None,
                               list_of_models=None):

        all_results = {}
        for feature_id, column_id in zip(features_id, columns_id):
            features = {}
            for rule in list_of_rules:
                features[rule] = self.import_feature(feature_id, column_id=column_id, rule=rule)

            results = {}
            for model in list_of_models:
                results[model] = {}
                for rule in list_of_rules:
                    feature = features[rule]
                    total_value = 0
                    ctr = 0
                    for instance in feature:
                        if model.lower() in instance.lower():
                            total_value += feature[instance]
                            ctr += 1
                    if ctr == 0:
                        avg_value = -1
                    else:
                        avg_value = round(total_value / ctr, 2)
                    results[model][rule] = avg_value
            all_results[f'{feature_id}_{column_id}'] = results

        for rule in list_of_rules:
            print(rule, end=" ")
            for model in list_of_models:
                print(f'&', end=" ")
                for feature_id, column_id in zip(features_id, columns_id):
                    print(f'{all_results[f"{feature_id}_{column_id}"][model][rule]}', end=" \ ")
            print("\\\\ \\midrule")

    def add_folders_to_experiment(self) -> None:

        dirs = ["experiments", "images", "trash"]
        for dir in dirs:
            if not os.path.isdir(dir):
                os.mkdir(os.path.join(os.getcwd(), dir))

        if not os.path.isdir(os.path.join(os.getcwd(), "experiments", self.experiment_id)):
            os.mkdir(os.path.join(os.getcwd(), "experiments", self.experiment_id))

        list_of_folders = ['distances',
                           'features',
                           'coordinates',
                           'elections',
                           'matrices']

        for folder_name in list_of_folders:
            if not os.path.isdir(os.path.join(os.getcwd(), "experiments",
                                              self.experiment_id, folder_name)):
                os.mkdir(os.path.join(os.getcwd(), "experiments",
                                      self.experiment_id, folder_name))

        path = os.path.join(os.getcwd(), "experiments", self.experiment_id, "map.csv")
        if not os.path.exists(path):
            with open(path, 'w') as file_csv:
                file_csv.write("size;num_candidates;num_voters;culture_id;params;color;alpha;"
                               "label;marker;show;path")
                file_csv.write("1;50;200;ic;{'p': 0.5};black;0.75;IC 0.5;*;process_id;{}")


    def convert_pb_to_app(self, **kwargs):
        pabulib.convert_pb_to_app(self, **kwargs)

# # # # # # # # # # # # # # # #
# LAST CLEANUP ON: 22.10.2021 #
# # # # # # # # # # # # # # # #



#!/usr/bin/env python
import logging

import numpy as np

from mapel.core.features.mallows import mallows_votes


def generate_mallows_urn_votes(num_voters: int = None,
                       num_candidates: int = None,
                       alpha: float = None,
                       phi: float = None,
                       **kwargs) -> np.ndarray:

    if alpha < 0:
        logging.warning(f'Incorrect value of alpha: {alpha}. Value should be >=0')

    votes = np.zeros([num_voters, num_candidates], dtype=int)
    urn_size = 1.
    for j in range(num_voters):
        rho = np.random.uniform(0, urn_size)
        if rho <= 1.:
            votes[j] = np.random.permutation(num_candidates)
        else:
            votes[j] = votes[np.random.randint(0, j)]
        urn_size += alpha

    votes = mallows_votes(votes, phi=phi)
    return votes



import csv
from collections import Counter

from mapel.core.glossary import *
from mapel.core.utils import *


def export_votes_to_file(election,
                         culture_id: str,
                         num_candidates: int,
                         num_voters: int,
                         params: dict,
                         path,
                         ballot_type,
                         votes=None,
                         is_aggregated=True):
    """
    Exports votes to a file.

    Parameters
    ----------
        election
            Election.
        culture_id: str
            Culture id.
        num_candidates : int
            Number of Candidates
        num_voters : int
            Number of Voters
        params : int
            Model params to be exported.
        path:
            Path to the place in which the file should be stored.
        ballot_type:
            Ballot type.
        votes:
            Votes.
        is_aggregated : bool
            If True then votes are stored in aggregated way.
    """
    if votes is None:
        votes = election.votes

    if params is None:
        params = {}

    with open(path, 'w') as file_:
        if ballot_type == 'ordinal':
            file_.write(f'# FILE NAME: {election.election_id}.soc\n')
            file_.write(f'# DATA TYPE: soc \n')
        elif ballot_type == 'approval':
            file_.write(f'# FILE NAME: {election.election_id}.app\n')
            file_.write(f'# DATA TYPE: app \n')
        file_.write(f'# CULTURE ID: {culture_id} \n')
        file_.write(f'# PARAMS: {str(params)} \n')
        file_.write(f'# NUMBER ALTERNATIVES: {num_candidates} \n')
        file_.write(f'# NUMBER VOTERS: {num_voters} \n')

        if is_aggregated:

            c = Counter(map(tuple, votes))
            counted_votes = [[count, list(row)] for row, count in c.items()]
            counted_votes = sorted(counted_votes, reverse=True)

            if ballot_type == 'approval':
                for i in range(len(counted_votes)):
                    file_.write(str(counted_votes[i][0]) + ': {')
                    for j in range(len(counted_votes[i][1])):
                        file_.write(str(int(counted_votes[i][1][j])))
                        if j < len(counted_votes[i][1]) - 1:
                            file_.write(", ")
                    file_.write("}\n")

            elif ballot_type == 'ordinal':
                for i in range(len(counted_votes)):
                    file_.write(str(counted_votes[i][0]) + ': ')
                    for j in range(len(counted_votes[i][1])):
                        file_.write(str(int(counted_votes[i][1][j])))
                        if j < len(counted_votes[i][1]) - 1:
                            file_.write(", ")
                    file_.write("\n")
        else:

            file_.write(str(num_voters) + ', ' + str(num_voters) + ', ' +
                        str(num_voters) + "\n")

            if ballot_type == 'approval':
                for i in range(len(votes)):
                    file_.write('1: {')
                    for j in range(len(votes[i])):
                        file_.write(str(int(list(votes[i])[j])))
                        if j < len(votes[i]) - 1:
                            file_.write(", ")
                    file_.write("}\n")

            elif ballot_type == 'ordinal':
                for i in range(len(votes)):
                    file_.write('1: ')
                    for j in range(len(votes[i])):
                        file_.write(str(int(votes[i][j])))
                        if j < len(votes[i]) - 1:
                            file_.write(", ")
                    file_.write("\n")


def export_approval_election(election,
                             is_aggregated: bool = True):
    """
    Exports approval election in an .app file

    Parameters
    ----------
        election
            Election.
        is_aggregated : bool
            If True then votes are stored in aggregated way.
    """
    path_to_folder = os.path.join(os.getcwd(), "experiments", election.experiment_id, "elections")
    make_folder_if_do_not_exist(path_to_folder)
    path_to_file = os.path.join(path_to_folder, f'{election.election_id}.app')

    if election.culture_id in APPROVAL_FAKE_MODELS:
        file_ = open(path_to_file, 'w')
        file_.write(f'$ {election.culture_id} {election.params} \n')
        file_.write(str(election.num_candidates) + '\n')
        file_.write(str(election.num_voters) + '\n')
        file_.close()

    else:
        export_votes_to_file(election,
                             election.culture_id,
                             election.num_candidates,
                             election.num_voters,
                             election.params,
                             path_to_file,
                             ballot_type='approval',
                             votes=election.votes,
                             is_aggregated=is_aggregated)


def export_ordinal_election(election,
                            is_aggregated: bool = True):
    """
    Exports ordinal election to a .soc file

    Parameters
    ----------
        election
            Election.
        is_aggregated : bool
            If True then votes are stored in aggregated way.
    """

    path_to_folder = os.path.join(os.getcwd(), "experiments", election.experiment_id, "elections")
    make_folder_if_do_not_exist(path_to_folder)
    path_to_file = os.path.join(path_to_folder, f'{election.election_id}.soc')

    if election.culture_id in LIST_OF_FAKE_MODELS:
        file_ = open(path_to_file, 'w')
        file_.write(f'$ {election.culture_id} {election.params} \n')
        file_.write(str(election.num_candidates) + '\n')
        file_.write(str(election.num_voters) + '\n')
        file_.close()
    else:
        export_votes_to_file(election,
                             election.culture_id,
                             election.num_candidates,
                             election.num_voters,
                             election.params,
                             path_to_file,
                             ballot_type='ordinal',
                             votes=election.votes,
                             is_aggregated=is_aggregated)


def export_distances(experiment,
                     object_type: str = 'vote',
                     length: int = None):
    """
    Exports distances to a csv file.

    Parameters
    ----------
        experiment
            Experiment.
        object_type : str
            Object type.
        length : int
            Number of elections.
    """

    file_name = f'{experiment.election_id}_{object_type}.csv'
    path = os.path.join(os.getcwd(), "experiments", experiment.experiment_id, "distances",
                        file_name)
    with open(path, 'w', newline='') as csv_file:
        writer = csv.writer(csv_file, delimiter=';')
        writer.writerow(["v1", "v2", "distance"])
        for v1 in range(length):
            for v2 in range(length):
                distance = str(experiment.distances[object_type][v1][v2])
                writer.writerow([v1, v2, distance])


def export_coordinates(experiment,
                       object_type: str = 'vote',
                       length: int = None):
    """
    Exports coordinates to a csv file.

    Parameters
    ----------
        experiment
            Experiment.
        object_type : str
            Object type.
        length : int
            Number of elections.
    """

    file_name = f'{experiment.election_id}_{object_type}.csv'
    path = os.path.join(os.getcwd(), "experiments", experiment.experiment_id, "coordinates",
                        file_name)
    with open(path, 'w', newline='') as csv_file:
        writer = csv.writer(csv_file, delimiter=';')
        writer.writerow(["vote_id", "x", "y"])
        for vote_id in range(length):
            x = str(experiment.coordinates[object_type][vote_id][0])
            y = str(experiment.coordinates[object_type][vote_id][1])
            writer.writerow([vote_id, x, y])



#!/usr/bin/env python


import itertools
import struct

import numpy as np


def banzhaf(rgWeights, fpThold=0.5, normalize=True):
    """ Compute Banzhaf power index """
    n = len(rgWeights)
    wSum = sum(rgWeights)
    wThold = fpThold * wSum
    rgWeights = np.array(rgWeights)
    cDecisive = np.zeros(n, dtype=np.uint64)
    for bitmask in range(0, 2**n-1):
        w = rgWeights * np.unpackbits(np.uint8(struct.unpack('B' * (8), np.uint64(bitmask))),
                                      bitorder='little')[:n]
        wSum = sum(w)
        if wSum >= wThold:
            cDecisive = np.add(cDecisive, (w > (wSum - wThold)))
    phi = cDecisive / (2**n) * 2
    if normalize:
        return phi / sum(phi)
    else:
        return phi


def shapley(rgWeights, fpThold=0.5):
    """ Compute Shapley-Shubik power index """
    n = len(rgWeights)
    wSum = sum(rgWeights)
    wThold = fpThold * wSum
    rgWeights = np.array(rgWeights)
    cDecisive = np.zeros(n, dtype=np.uint64)
    for perm in itertools.permutations(range(n)):
        w = rgWeights[list(perm)]
        dec = 0
        wSum = w[dec]
        while wSum < wThold:
            dec += 1
            wSum += w[dec]
        cDecisive[perm[dec]] += 1
    return cDecisive / sum(cDecisive)




import numpy as np
from scipy.stats import gamma
import random as rand
import mapel.core.features.mallows as mallows
from mapel.core.glossary import *


# Ordinal
def update_params_ordinal_mallows(params):
    if 'phi' in params and type(params['phi']) is list:
        params['phi'] = np.random.uniform(low=params['phi'][0], high=params['phi'][1])
    elif 'phi' not in params:
        params['phi'] = np.random.random()


def update_params_ordinal_norm_mallows(params, num_candidates):
    if 'normphi' not in params:
        params['normphi'] = np.random.random()
    params['phi'] = mallows.phi_from_normphi(num_candidates, normphi=params['normphi'])
    if 'weight' not in params:
        params['weight'] = 0.


def update_params_ordinal_urn_model(params):
    if 'alpha' not in params:
        params['alpha'] = gamma.rvs(0.8)


def update_params_ordinal_mallows_matrix_path(params, num_candidates):
    params['normphi'] = params['alpha']
    params['phi'] = mallows.phi_from_normphi(num_candidates, normphi=params['normphi'])


def update_params_ordinal_mallows_triangle(params, num_candidates):
    params['normphi'] = 1 - np.sqrt(np.random.uniform())
    params['phi'] = mallows.phi_from_normphi(num_candidates, normphi=params['normphi'])
    params['weight'] = np.random.uniform(0, 0.5)
    params['alpha'] = params['normphi']
    params['tint'] = params['weight']  # for tint on plots


def update_params_ordinal_alpha(printing_params):
    if 'alpha' not in printing_params:
        printing_params['alpha'] = None
    elif type(printing_params['alpha']) is list:
        printing_params['alpha'] = np.random.uniform(low=printing_params['alpha'][0],
                                                     high=printing_params['alpha'][1])


def update_params_ordinal_preflib(params, model_id):
    # list of IDs larger than 10
    folder = ''
    if model_id == 'irish':
        folder = 'irish_s1'
        # folder = 'irish_f'
        ids = [1, 3]
    elif model_id == 'glasgow':
        folder = 'glasgow_s1'
        # folder = 'glasgow_f'
        ids = [2, 3, 4, 5, 6, 7, 8, 9, 11, 13, 16, 19, 21]
    elif model_id == 'formula':
        folder = 'formula_s1'
        # 17 races or more
        ids = [17, 35, 37, 40, 41, 42, 44, 45, 46, 47, 48]
    elif model_id == 'skate':
        folder = 'skate_ic'
        # 9 judges
        ids = [1, 2, 3, 4, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
               25, 26, 27, 28, 29, 30, 31, 32, 33, 34,
               35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 47, 48]
    elif model_id == 'sushi':
        folder = 'sushi_ff'
        ids = [1]
    elif model_id == 'grenoble':
        folder = 'grenoble_ff'
        ids = [1]
    elif model_id == 'tshirt':
        folder = 'tshirt_ff'
        ids = [1]
    elif model_id == 'cities_survey':
        folder = 'cities_survey_s1'
        ids = [1, 2]
    elif model_id == 'aspen':
        folder = 'aspen_s1'
        ids = [1]
    elif model_id == 'marble':
        folder = 'marble_ff'
        ids = [1, 2, 3, 4, 5]
    elif model_id == 'cycling_tdf':
        folder = 'cycling_tdf_s1'
        # ids = [e for e in range(1, 69+1)]
        selection_method = 'random'
        ids = [14, 15, 16, 17, 18, 19, 20, 21, 23, 24, 25, 26]
    elif model_id == 'cycling_gdi':
        folder = 'cycling_gdi_s1'
        ids = [i for i in range(2, 23 + 1)]
    elif model_id == 'ers':
        folder = 'ers_s1'
        # folder = 'ers_f'
        # 500 voters or more
        ids = [3, 9, 23, 31, 32, 33, 36, 38, 40, 68, 77, 79, 80]
    elif model_id == 'ice_races':
        folder = 'ice_races_s1'
        # 80 voters or more
        ids = [4, 5, 8, 9, 15, 20, 23, 24, 31, 34, 35, 37, 43, 44, 49]
    else:
        ids = []

    if 'id' not in params:
        params['id'] = rand.choices(ids, k=1)[0]

    params['folder'] = folder


def update_params_ordinal(params, printing_params, variable, culture_id, num_candidates):
    if variable is not None:
        printing_params['alpha'] = params[variable]
        printing_params['variable'] = variable
    else:
        if culture_id.lower() == 'mallows':
            update_params_ordinal_mallows(params)
            printing_params['alpha'] = params['phi']
        elif 'norm_mallows' in culture_id.lower() or 'norm-mallows' in culture_id.lower() \
                or 'mallows_urn' in culture_id.lower():
            update_params_ordinal_norm_mallows(params, num_candidates)
            printing_params['alpha'] = params['normphi']
        elif 'urn' in culture_id.lower():
            update_params_ordinal_urn_model(params)
            printing_params['alpha'] = params['alpha']
        elif culture_id.lower() == 'mallows_matrix_path':
            update_params_ordinal_mallows_matrix_path(params, num_candidates)
        elif culture_id.lower() == 'mallows_triangle':
            update_params_ordinal_mallows_triangle(params, num_candidates)
        elif culture_id.lower() in LIST_OF_PREFLIB_MODELS:
            update_params_ordinal_preflib(params, culture_id)
        update_params_ordinal_alpha(printing_params)
    return params, printing_params


# Approval #
def update_params_approval_alpha(printing_params):
    if 'alpha' not in printing_params:
        printing_params['alpha'] = 1
    elif type(printing_params['alpha']) is list:
        printing_params['alpha'] = np.random.uniform(low=printing_params['alpha'][0],
                                                     high=printing_params['alpha'][1])


def update_params_approval_p(params):
    if 'p' not in params:
        params['p'] = np.random.rand()
    elif type(params['p']) is list:
        params['p'] = np.random.uniform(low=params['p'][0], high=params['p'][1])


def update_params_approval_resampling(params, printing_params):
    if 'phi' in params and type(params['phi']) is list:
        params['phi'] = np.random.uniform(low=params['phi'][0], high=params['phi'][1])
    elif 'phi' not in params:
        params['phi'] = np.random.random()
    printing_params['alpha'] = params['phi']

    if 'p' in params and type(params['p']) is list:
        params['p'] = np.random.uniform(low=params['p'][0], high=params['p'][1])
    elif 'p' not in params:
        params['p'] = np.random.random()


def update_params_approval_disjoint(params, printing_params):
    if 'phi' in params and type(params['phi']) is list:
        params['phi'] = np.random.uniform(low=params['phi'][0], high=params['phi'][1])
    elif 'phi' not in params:
        params['phi'] = np.random.random()
    printing_params['alpha'] = params['phi']

    if 'p' in params and type(params['p']) is list:
        params['p'] = np.random.uniform(low=params['p'][0], high=params['p'][1])
    elif 'p' not in params:
        params['p'] = np.random.random() / params['g']


def update_params_approval(params, printing_params, variable, culture_id, num_candidates):
    printing_params['alpha'] = 0
    if variable is not None:
        if culture_id in APPROVAL_MODELS:
            update_params_approval_p(params)
        printing_params['alpha'] = params[variable]
        printing_params['variable'] = variable
        del params['variable']
    else:
        if culture_id.lower() == 'resampling':
            update_params_approval_resampling(params, printing_params)
        elif culture_id.lower() == 'disjoint':
            update_params_approval_disjoint(params, printing_params)
        elif culture_id in APPROVAL_MODELS:
            update_params_approval_p(params)
        update_params_approval_alpha(printing_params)

    if 'p' in params and culture_id in ['empty',
                                        'full',
                                        'euclidean',
                                        'urn_partylist']:
        del params['p']

    return params, printing_params


def get_params_for_crate(j):
    base = []
    my_size = 10
    # with_edge
    for p in range(my_size):
        for q in range(my_size):
            for r in range(my_size):
                a = p / (my_size - 1)
                b = q / (my_size - 1)
                c = r / (my_size - 1)
                d = 1 - a - b - c
                tmp = [a, b, c, d]
                if d >= 0 and sum(tmp) == 1:
                    base.append(tmp)
    params = {'alpha': base[j]}
    return params


def get_params_for_paths(family, j, extremes=False):
    path = family.path

    variable = path['variable']

    if 'extremes' in path:
        extremes = path['extremes']

    params = {'variable': variable}
    if extremes:
        params[variable] = j / (family.size - 1)
    elif not extremes:
        params[variable] = (j + 1) / (family.size + 1)

    if 'scale' in path:
        params[variable] *= path['scale']

    if 'start' in path:
        params[variable] += path['start']
    else:
        path['start'] = 0.

    if 'step' in path:
        params[variable] = path['start'] + j * path['step']

    return params, variable


from setuptools import setup 
from pybind11.setup_helpers import Pybind11Extension, build_ext

cpp_dist_module = Pybind11Extension('mapel.elections.distances.cppdistances',
['src/mapel/elections/distances/cppdistances.cpp'], extra_compile_args=["-Ofast"])

setup(
  ext_modules = [cpp_dist_module],
  cmdclass={"build_ext": build_ext},
)



"""
Script which generates map.csv file for experiment with syntetic elections of varying number of candidates.
The statictical culture is always displayed in the same color with varying levels of transparency depending on the number of candidates - by default lighter shade for lowe number of candidates
This function can be run
Optional parameters (bools by default true unless stated otherwise):
    candidates: set to [8, 9, 10, 11, 12] by default, meaning
    alphas: transparency for the elections with number of candidates on corresponding indeces in the candidates list
    family_size: number of elections per family
    num_voters: number of voters per election
    file_name: file to which the output should be saved
    incl_ic: bool whether IC culture should be included
    incl_iac: bool whether IAC culture should be included
    incl_conitzer: bool whether conitzer culture should be included
    incl_walsh: bool whether walsh culture should be included
    incl_sc: bool whether single crossing culture should be included
    incl_mallows_default: bool whether Mallows with default values of phi (0.3, 0.5, 0.7) should be included
    incl_mallows_custom: bool whether Mallows with custom values of phi should be included - set by default to False, if set to True, then list of phi also needs to be added
    mallows_params_phi: custom phis that will be added into the csv file
    mallows_params_colors: colors of corresponding custom phis for Mallows elections
    incl_euclidean: bool whether euclidean cultures should be included
    incl_extremes: bool whether UN, AN, ID and IDAN should be included

"""

def generate_mapcsv_contents(*args, **kwargs):
    candidates = [8, 9, 10, 11, 12]
    alphas = [0.2, 0.4, 0.6, 0.8, 1]
    family_size = 10
    urn_family_size = 10
    mallows_family_size = 10
    num_voters = 100
    file_name = "map.txt"
    incl_ic = True
    incl_iac = False
    incl_urn = True
    incl_conitzer = True
    incl_walsh = True
    incl_sc = True
    incl_spoc = True
    incl_mallows_default = True
    incl_mallows_custom = False
    mallows_params_phi = None
    mallows_params_colors = None
    incl_euclidean = True
    incl_extremes = True

    for key, val in kwargs.items():
        if key == "candidates":
            candidates = val
        elif key == "alphas":
            alphas = val
        elif key == "family_size":
            family_size = val
        elif key == "num_voters":
            num_voters = val
        elif key == "file_name":
            file_name = val
        elif key == "incl_ic":
            incl_ic = val
        elif key == "incl_iac":
            incl_iac = val
        elif key == "incl_conitzer":
            incl_conitzer = val
        elif key == "incl_walsh":
            incl_walsh = val
        elif key == "incl_sc":
            incl_sc = val
        elif key == "incl_spoc":
            incl_spoc = val
        elif key == "incl_mallows_default":
            incl_mallows_default = val
        elif key == "incl_mallows_custom":
            incl_mallows_default = False
            incl_mallows_custom = True
        elif key == "mallows_params_phi":
            mallows_params_phi = val
        elif key == "mallows_params_colors":
            mallows_params_colors = val
        elif key == "incl_euclidean":
            incl_euclidean = val
        elif key == "incl_extremes":
            incl_extremes = val
        else:
            print("Optional argument \"", key, "\" does not exist.")

    if incl_mallows_custom is True and (mallows_params_phi is None or mallows_params_colors is None):
        incl_mallows_custom = False
        print("Missing parameter values for custom mallows, so no custom mallows will be added.")

    file = open(file_name, 'w')
    file.write("size;num_candidates;num_voters;culture_id;params;color;alpha;family_id;marker;path;label\n")

    for i in range(0, len(candidates)):
        if incl_ic:
            ic_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "ic" + ";" + "{}" + ";" + "black" + ";" + str(
                alphas[i]) + ";" + "ic" + str(candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "IC \n"
            file.write(ic_string)

        if incl_iac:
            iac_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "iac" + ";" + "{}" + ";" + "lavender" + ";" + str(
                alphas[i]) + ";" + "iac" + str(candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "IAC \n"
            file.write(iac_string)

        if incl_urn:
            urn_alpha = [0.01, 0.02, 0.05, 0.1, 0.5]
            urn_alpha_color = ['palegoldenrod', 'yellow', 'gold', 'orange', 'red']
            for j in range(0, len(urn_alpha)):
                urn_string = str(urn_family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "urn" + ";" + "{'alpha': " + str(urn_alpha[j]) + "}" + ";" + str(urn_alpha_color[j]) + ";" + str(
                alphas[i]) + ";" + "urn" + str(candidates[i]) + '-' + str(urn_alpha[j]) + ";" + "o" + ";" + "{}" + ";" + "Urn-" + str(urn_alpha[j]) + "\n"
                file.write(urn_string)

        if incl_conitzer:
            conitzer_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "conitzer" + ";" + "{}" + ";" + "saddlebrown" + ";" + str(
                alphas[i]) + ";" + "conitzer" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "SP-Conitzer \n"
            file.write(conitzer_string)

        if incl_walsh:
            walsh_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "walsh" + ";" + "{}" + ";" + "olive" + ";" + str(
                alphas[i]) + ";" + "walsh" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "SP-Walsh \n"
            file.write(walsh_string)

        if incl_sc:
            singlecrossing_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "single-crossing" + ";" + "{}" + ";" + "mediumorchid" + ";" + str(
                alphas[i]) + ";" + "single-crossing" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "SC \n"
            file.write(singlecrossing_string)

        if incl_mallows_default:
            phis = [0.001, 0.01, 0.1, 0.5, 0.75, 0.99]
            phis_colors = ['paleturquoise', 'skyblue', 'dodgerblue', 'cornflowerblue', 'blue', 'navy']
            for j in range(0, len(phis)):
                mallows_string = str(mallows_family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "norm-mallows" + ";" + "{'normphi':" + str(phis[j]) + "}" + ";" + phis_colors[j] + ";" + str(
                alphas[i]) + ";" + "norm-mallows-" + str(candidates[i]) + '-' + str(phis[j]) + ";" + "o" + ";" + "{}" + ";" + "Norm-Mallows-" + str(phis[j]) + "\n"
                file.write(mallows_string)

        if incl_mallows_custom:
            for j in range(0, len(mallows_params_phi)):
                mallows_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                    num_voters) + ";" + "norm-mallows" + ";" + "{'normphi': " + str(mallows_params_phi[j]) +"}" + ";" + mallows_params_colors[j] + ";" + str(
                    alphas[i]) + ";" + "norm-mallows" + str(candidates[i]) + str(
                    mallows_params_phi[j]) + ";" + "x" + ";" + "{}" + ";" + "Norm-Mallows-" +str(mallows_params_phi[j]) +"\n"
                file.write(mallows_string)

        if incl_spoc:
            spoc_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "spoc_conitzer" + ";" + "{}" + ";" + "firebrick" + ";" + str(
                alphas[i]) + ";" + "spoc_conitzer" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "SPOC \n"
            file.write(spoc_string)


        if incl_euclidean:
            euclidean_1_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "euclidean" + ";" + "{'dim': 1, 'space': 'uniform'}" + ";" + "lime" + ";" + str(
                alphas[i]) + ";" + "Interval" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "1D Interval \n"
            euclidean_2_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "euclidean" + ";" + "{'dim': 2, 'space': 'uniform'}" + ";" + "green" + ";" + str(
                alphas[i]) + ";" + "Square" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "Square \n"
            euclidean_3_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "euclidean" + ";" + "{'dim': 3, 'space': 'uniform'}" + ";" + "forestgreen" + ";" + str(
                alphas[i]) + ";" + "Cube" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "3-Cube \n"
            euclidean_5_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "euclidean" + ";" + "{'dim': 5, 'space': 'uniform'}" + ";" + "palegreen" + ";" + str(
                alphas[i]) + ";" + "5-Cube" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "5-Cube \n"
            euclidean_10_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "euclidean" + ";" + "{'dim': 10, 'space': 'uniform'}" + ";" + "yellowgreen" + ";" + str(
                alphas[i]) + ";" + "10-Cube" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "10-Cube \n"
            euclidean_2_sphere_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "euclidean" + ";" + "{'dim': 2, 'space': 'sphere'}" + ";" + "deeppink" + ";" + str(
                alphas[i]) + ";" + "Circle" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "Circle \n"
            euclidean_3_sphere_string = str(family_size) + ";" + str(candidates[i]) + ";" + str(
                num_voters) + ";" + "euclidean" + ";" + "{'dim': 3, 'space': 'sphere'}" + ";" + "lightpink" + ";" + str(
                alphas[i]) + ";" + "Sphere" + str(
                candidates[i]) + ";" + "o" + ";" + "{}" + ";" + "Sphere \n"
            file.write(euclidean_1_string)
            file.write(euclidean_2_string)
            file.write(euclidean_3_string)
            file.write(euclidean_5_string)
            file.write(euclidean_10_string)
            file.write(euclidean_2_sphere_string)
            file.write(euclidean_3_sphere_string)

    if incl_extremes:
        st_string = '1' + ';' + str(15) + ';' + '50' + ';real_stratification;{};black;' + str(1) + ';ST;x;{};ST \n'
        un_string = '1' + ';' + str(15) + ';' + '50' + ';un_from_matrix;{};black;' + str(1) + ';UN;x;{};UN \n'
        id_string = '1' + ';' + str(15) + ';' + '50' + ';real_identity;{};rosybrown;' + str(1) + ';ID;x;{};ID \n'
        an_string = '1' + ';' + str(15) + ';' + '50' + ';real_antagonism;{};royalblue;' + str(1) + ';AN;x;{};AN \n'
        idan_string = '4' + ';' + str(15) + ';' + '50' + ';idan_part;{};midnightblue;' + str(1) + ';IDAN;x;{\'variable\' : \'part_share\'};IDAN \n'
        file.write(st_string)
        file.write(un_string)
        file.write(id_string)
        file.write(an_string)
        file.write(idan_string)

    file.close()


if __name__ == "__main__":
    generate_mapcsv_contents()



from prefsampling.ordinal.euclidean import EuclideanSpace
import prefsampling.ordinal as pref_ordinal


def euclidean_mask(num_voters=None,
                   num_candidates=None,
                   space=None,
                   **kwargs):
    if type(space) is str:
        if space.lower() == 'uniform':
            space = EuclideanSpace.UNIFORM
        elif space.lower() == 'sphere':
            space = EuclideanSpace.SPHERE
        elif space.lower() == 'gaussian':
            space = EuclideanSpace.GAUSSIAN
        elif space.lower() == 'ball':
            space = EuclideanSpace.BALL

    if space is None:
        space = EuclideanSpace.UNIFORM

    return pref_ordinal.euclidean(num_voters=num_voters,
                                  num_candidates=num_candidates,
                                  space=space,
                                  **kwargs)



import prefsampling.approval as pref_approval
import prefsampling.ordinal as pref_ordinal


def truncated_urn_mask(num_voters=None,
                       num_candidates=None,
                       p=None,
                       alpha=None,
                       **kwargs):
    return pref_approval.truncated_ordinal(num_voters=num_voters,
                                           num_candidates=num_candidates,
                                           p=p,
                                           ordinal_sampler=pref_ordinal.urn,
                                           ordinal_sampler_parameters={'alpha': alpha})




import queue
import numpy as np


def print_tree(root):
    print(root.election_id_init, root.election_id, root.value)
    for child in root.children:
        print_tree(child)


class Node:

    def __init__(self, election_id):
        self.election_id = election_id
        self.election_id_init = election_id
        self.parent = None
        self.children = []
        self.leaf = True
        self.value = None

    def __str__(self):
        return self.election_id

    def add_child(self, child, value):
        child.parent = self
        child.value = value
        self.children.append(child)
        self.leaf = False


def _prepare_tree(vector):

    num_leaves = len(vector)

    root = Node('root')
    ctr = 0

    q = queue.Queue()
    q.put(root)

    while q.qsize() * 2 < num_leaves:
        tmp_root = q.get()
        for _ in range(2):
            inner_node = Node('v' + str(ctr))
            tmp_root.add_child(inner_node, 0)
            q.put(inner_node)
            ctr += 1

    ctr = 0
    while ctr < num_leaves:
        tmp_root = q.get()
        for _ in range(2):
            node = Node('x' + str(ctr))
            tmp_root.add_child(node, vector[ctr])
            ctr += 1

    return root


def _run_over_tree(node):
    if not node.children[0].leaf:
        for child in node.children:
            _run_over_tree(child)

    # randomized_compare
    a = node.children[0].value
    b = node.children[1].value

    if a is None:
        node.value = b
        node.election_id = node.children[1].election_id

    elif b is None:
        node.value = a
        node.election_id = node.children[0].election_id

    else:

        if a + b >= 1:
            node.value = a + b - 1
            if np.random.random() > (1-a) / ((1-a)+(1-b)+0.00001):
                node.children[0].value = 1
                node.election_id = node.children[1].election_id
            else:
                node.children[1].value = 1
                node.election_id = node.children[0].election_id

        else:
            node.value = a + b
            if np.random.random() > a / (a+b+0.00001):
                node.children[0].value = 0
                node.election_id = node.children[1].election_id
            else:
                node.children[1].value = 0
                node.election_id = node.children[0].election_id

        if a+b == 1.:
            node.value = None


def _get_final_values_from_tree(node, values):
    if node.value is not None and node.election_id not in values:
        values[node.election_id] = int(node.value)
    for child in node.children:
        _get_final_values_from_tree(child, values)


def approx_rand_tree(values):
    root = _prepare_tree(values)
    _run_over_tree(root)
    values = {}
    _get_final_values_from_tree(root, values)
    return values

